import tensorflow as tf
import opt_einsum as oe

class WeightedContractionLayer(tf.keras.layers.Layer):
    '''
    Given the input tensor, this layer performs an einsum contraction with learnable weights on the input (i.e. the tensor the input is einsummed with is learnable)
    '''
    def __init__(self, contraction_expression = None, new_weight_dims = [], contraction_expression_contains_batch_index = False, softmax_weights = True, **kwargs):
        '''
        Init arguments
        
        contraction_expression: str. Must be a valid opt_einsum contraction expression. First operand represents the weights to be created. Unlike typical einsum, if you have ellipses in the input tensor and you want to create the appropriately sized weight tensor for those axes, you MUST use an ellipsis in the weight tensor indices as well. Do NOT USE Theta as a tensor index - this is used to handle ellipses.
        new_weight_dims: List of the size of new dimensions introduced in the weight tensor in the order they appear. For example, the contraction 'mk,jkl->ml' has a new axis m for the weights, so the user must specify its size. If an int is supplied, it will be automatically converted to a 1 element long list.
        contraction_expression_contains_batch_index: Boolean. Every tensor keras works with has 1 batch dimension as the 0th dim. Set this to False if your contraction_expression does not account for this - the init will take care of it.
        softmax_weights: Boolean. If set to true, the weights will be softmaxed before the contraction is performed.
        '''
        if not isinstance(contraction_expression, str):
            raise(TypeError('Contraction expression must be a string valid for np.einsum'))
        
        if isinstance(new_weight_dims, int):
            new_weight_dims = [new_weight_dims]
        
        self.softmax_weights = softmax_weights
        
        #parse contraction_expression
        self.contraction_expression = contraction_expression
        input_expressions, self.output_expression = contraction_expression.replace(' ', '').split('->')
        self.weight_expression, self.input_expression = input_expressions.replace('...', 'Θ').split(',')
        self.new_weight_dims = new_weight_dims
        
        #find which indices are shared and which aren't between the weights and the input
        self.shared_indices = list(set(self.input_expression) & set(self.weight_expression)) #indices to be shared between the weight tensor and the input
        self.new_indices = list(set(self.weight_expression) - set(self.input_expression)) #new indices to be created that don't correspond to a dim in the weight tensor
        
        #infer new axis weight lengths
        self.weight_shape = []
        if (new_weight_dims == []) and (self.new_indices != []):
            raise(ValueError('New weight dimensions must be not None if new dimensions are being created'))
        elif len(self.new_indices) != len(new_weight_dims):
            raise(ValueError('Supply exactly one size for every new dimension'))
        
        for i in range(len(self.weight_expression)):
            if self.weight_expression[i] in self.new_indices:
                self.weight_shape.append(new_weight_dims.pop(0))
            else:
                self.weight_shape.append(None)

        #correct contraction expression for batch index if not supplied
        if not contraction_expression_contains_batch_index:
            self.contraction_expression = self.contraction_expression[:self.contraction_expression.find('->')+2] + 'z' + self.contraction_expression[self.contraction_expression.find('->')+2:]
            self.contraction_expression = self.contraction_expression[:self.contraction_expression.find(',')+1] + 'z' + self.contraction_expression[self.contraction_expression.find(',')+1:]
            
        super().__init__(**kwargs)
    
    def build(self, input_shape):
        #infer shared axis weight lengths
        input_shape = input_shape[1:]
        for shared_index in self.shared_indices:
            input_dim_number = self.input_expression.find(shared_index)
            weight_dim_number = self.weight_expression.find(shared_index)
            try:
                self.weight_shape[weight_dim_number] = int(input_shape[input_dim_number])
            except:
                raise(ValueError('Shared indices along which a contraction is performed must have predetermined lengths'))
        #create weights
        self.contraction_weights = self.add_weight(name = 'contraction_weights', shape = self.weight_shape, initializer='glorot_uniform', trainable=True)
        super().build(input_shape)
        
    def call(self, inp):
        if None in inp.shape: #This is ONLY necessary to not throw an exception during the call after __init__. DO NOT REMOVE.
            newshape = [1 if v is None else v for v in inp.shape]
            inp = tf.random.uniform(newshape, dtype = tf.keras.backend.floatx())
        #contract
        if self.softmax_weights:
            return oe.contract(self.contraction_expression, tf.nn.softmax(self.contraction_weights), inp, backend = 'tensorflow')
        else:
            return oe.contract(self.contraction_expression, self.contraction_weights, inp, backend = 'tensorflow')



from .MergeWithAttention import MergeWithAttention
from .SpatialPyramidPool import SpatialPyramidPool
from .Upsample import Upsample
from .WeightedContractionLayer import WeightedContractionLayer
#from .Scaling import Scaling
from .DeconvUpscale import *



import numpy as np
import copy
import tensorflow as tf
from collections.abc import Iterable, Iterator
import itertools



def poisson_matrix(m,n):
    '''
    Generates the matrix A to express the Poisson equation in the form Ax=b for an m-by-n grid
    
    The matrix returned shall be (m-2)*(n-2)-by-(m-2)*(n-2) in size
    
    YOU MUST RESHAPE THE RESULT FROM (i.e. solution = inv(poisson_matrix) * right_hand_size) FROM (...,(m-2)*(n-2)) TO (...,m-2,n-2) BY USING FORTRAN COLUMN-MAJOR ORDERING!!!!!!!!
    '''
    m = m-2
    n = n-2
    
    D = np.zeros((m,m), dtype = np.float64)
    i,j = np.indices(D.shape)
    D[i==j] = 4.0
    D[i==j-1] = -1.0
    D[i==j+1] = -1.0
    
    S = -np.eye(D.shape[0], dtype = np.float64)
    
    P = np.zeros((m*n,m*n), dtype = np.float64)
    ind = np.arange(0,m*(n+1), m)
    
    for i in range(len(ind)-1):
        P[ind[i]:ind[i+1], ind[i]:ind[i+1]] = D
        try:
            P[ind[i+1]:ind[i+2], ind[i]:ind[i+1]] = S
        except:
            pass
        try:
            P[ind[i-1]:ind[i], ind[i]:ind[i+1]] = S
        except:
            pass
    return P

    

def poisson_RHS(F, boundaries = None, h = None, rho = None):
    '''
    Generates the RHS vector b of a discretized Poisson problem in the form Ax=b.
    h = grid spacing. if not supplied, the RHS will be returned without the (delta x)^2 factor. necessary for nonuniform grid spacing.
    boundaries = dict containing entries 'top', 'bottom', 'right' and 'left' which correspond to the Dirichlet BCs at these boundaries. Each entry must be a vector of length m or n, where m and n are defined as in te function poisson_matrix
    F = an m by n matrix containing the RHS values of the Poisson equation
    
    (i.e. this function merely takes the BC information and the (structured) RHS array to provide the RHS for the matrix eq. form)
    '''
    #print(F)
    if isinstance(F, list):
        boundaries = F[1]
        try:
            h = F[2]
        except:
            pass
        try:
            rho = F[3]
        except:
            pass
        F = F[0]
    
    boundaries = copy.deepcopy(boundaries)
    F = copy.deepcopy(F)
    
    for key in boundaries.keys():
        if len(boundaries[key].shape) > 1:
            boundaries[key] = tf.squeeze(boundaries[key])
        if len(boundaries[key].shape) == 1:
            boundaries[key] = itertools.repeat(boundaries[key], F.shape[0])
    
    if (h is not None) and isinstance(h, float):
        h = itertools.repeat(h, F.shape[0])
    
    for i in range(F.shape[0]):
        if h is not None:
            try:
                dx = next(h)
            except:
                dx = h[i]
            F[i] = -dx**2 * F[i]
                
        
        if isinstance(boundaries['top'], Iterator):
            top = next(boundaries['top'])[1:-1] 
        else:
            top = boundaries['top'][i,1:-1]
            
        if isinstance(boundaries['bottom'], Iterator):
            bottom = next(boundaries['bottom'])[1:-1]
        else:
            bottom = boundaries['bottom'][i,1:-1]
            
        if isinstance(boundaries['left'], Iterator):
            left = next(boundaries['left'])[1:-1]
        else:
            left = boundaries['left'][i,1:-1]
            
        if isinstance(boundaries['right'], Iterator):
            right = next(boundaries['right'])[1:-1]
        else:
            right = boundaries['right'][i,1:-1]
            
        if rho is not None:
            top = top/rho[i,...,1:-1,1]
            bottom = bottom/rho[i,...,1:-1,-2]
            left = left/rho[i,...,1,1:-1]
            right = right/rho[i,...,-2,1:-1]
        
        F[i,...,1:-1,1] += bottom
        F[i,...,1:-1,-2] += top
        F[i,...,1,1:-1] += left
        F[i,...,-2,1:-1] += right

    return F[...,1:-1,1:-1].reshape(list(F[...,1:-1,1:-1].shape[:-2]) + [np.prod(F[...,1:-1,1:-1].shape[-2:])], order = 'C')
 

def cholesky_poisson_solve(rhses, boundaries, h, system_matrix = None, system_matrix_is_decomposed = False):
    '''
    Solves the Poisson equation for the given RHSes.
    
    rhses: tf.Tensor representing the RHS functions of the Poisson equation, defined across the last 2 dimensions
    boundaries: boundary conditions of the outputs; see poisson_RHS documentation
    h: grid spacing of the outputs
    system_matrix: Poisson equation LHS matrix generated by poisson_matrix OR its Cholesky decomposition computed by tf.linalg.cholesky. If the Cholesky decomposition is supplied, set system_matrix_is_decomposed to True.
    system_matrix_is_decomposed: Flag to declare if a provided system_matrix is already Cholesky decomposed. Setting this to True when the system_matrix isn't decomposed will lead to undefined behaviour. Ignored if system_matrix is None.
    
    Outputs a tf.Tensor of identical shape to rhses.
    
    Note: Not tested if this function works on CPU.
    '''
    try: #handle spurious 1 dimensions
        rhses = tf.squeeze(rhses, axis = 1)
    except:
        pass
    boundaries = copy.deepcopy(boundaries)
    for boundary in boundaries.keys():
        try:
            boundaries[boundary] = tf.squeeze(boundaries[boundary], axis = 1)
        except:
            pass
    
    #generate poisson matrix, or use the provided one
    if system_matrix == None:
        system_matrix = tf.cast(poisson_matrix(int(rhses.shape[-2]), int(rhses.shape[-1])), tf.keras.backend.floatx())
        system_matrix_is_decomposed = False
    #import pdb
    #pdb.set_trace()
    if not system_matrix_is_decomposed:
        system_matrix_chol = tf.expand_dims(tf.linalg.cholesky(system_matrix), axis=0)
    else:
        system_matrix_chol = system_matrix
    
    def chol(r):
        return tf.linalg.cholesky_solve(system_matrix_chol, tf.expand_dims(r, axis = 0))
    
    try:
        @tf.contrib.eager.defun
        def chol_solve(rhs_arr):
            return tf.map_fn(chol, rhs_arr)
    except:
        @tf.function
        def chol_solve(rhs_arr):
            return tf.map_fn(chol, rhs_arr)
    
    #put problem into Ax=b format
    try:
        rhs_vectors = tf.expand_dims(tf.transpose(tf.squeeze(poisson_RHS([np.array(rhses), boundaries, h])), (0,1)),axis=2)
    except:
        rhs_vectors = tf.expand_dims(tf.expand_dims(tf.squeeze(poisson_RHS([np.array(rhses), boundaries, h])),axis=1), axis=0)
    
    #solve and reshape
    z = tf.reshape(chol_solve(rhs_vectors), list(rhses.shape[:-2]) + [int(rhses.shape[-1])-2, int(rhses.shape[-2])-2])
    
    soln = np.zeros(rhses.shape, dtype = np.float64)
    soln[...,:,-1] = boundaries['top']
    soln[...,:,0] = boundaries['bottom']
    soln[...,0,:] = boundaries['left']
    soln[...,-1,:] = boundaries['right']
    soln[...,1:-1,1:-1] = z
    
    return tf.expand_dims(soln, axis = 1)




from .numerical import numerical_dataset, numerical_dataset_generator, generate_random_RHS, generate_random_boundaries, set_max_magnitude_in_batch
from .analytical import homogeneous_analytical_dataset, homogeneous_analytical_dataset_generator



import numpy as np
import tensorflow.compat.v1 as tf
from collections.abc import Iterable

class Upsample(tf.keras.layers.Layer):
    '''
    Layer that upsamples the input images with one of the upscaling functions provided in tf.image.ResizeMethod module
    '''
    def __init__(self, dim_args, data_format = 'channels_first', resize_method = tf.image.ResizeMethod.BICUBIC, align_corners = True, **kwargs):
        super().__init__(**kwargs)
        for val in dim_args:
            if not (isinstance(val,float) or isinstance(val,int)):
                raise(TypeError('Supply floats or ints as the dimension arguments!'))
        
        self.align_corners = align_corners
        #channels first or channels last
        self.data_format = data_format
        #function to be used from tf.image.ResizeMethod
        self.resize_method = resize_method
        #set upsample ratio
        self.dim_args = dim_args #floats will scale that dimension to int(old_dim * dim_args[i]). ints will scale straight to that size.
        #self.first_call = True
        if self.data_format == 'channels_first':
            self.data_dims = [2,3]
        elif self.data_format == 'channels_last':
            self.data_dims = [1,2]
            
    def build(self, input_shape):
        super().build(input_shape)
    
    def get_newshape(self, inp_shape):
        new_shape = [inp_shape[i] for i in self.data_dims]
        for i in range(len(self.dim_args)): 
            if isinstance(self.dim_args[i],int):
                new_shape[i] = self.dim_args[i]
            elif isinstance(self.dim_args[i],float):
                if inp_shape[self.data_dims[i]] == None:
                    new_shape[i] = None
                else:
                    new_shape[i] = int(inp_shape[self.data_dims[i]] * self.dim_args[i])
        return new_shape
    def compute_output_shape(self, input_shape):
        new_shape = self.get_newshape(input_shape)
        if self.data_format == 'channels_first':
            return tf.TensorShape(list(input_shape)[0:2] + new_shape)
        else:
            return tf.TensorShape([input_shape[0]] + new_shape + [input_shape[-1]])
        
    def call(self, inputs):
        if isinstance(inputs, list) or isinstance(inputs, tuple):
            set_unknown_dims_as = inputs[1]
            inputs = inputs[0]
            newshape = self.get_newshape(inputs.shape)
            if -1 in newshape:
                newshape = [newshape[i] if newshape[i] != -1 else set_unknown_dims_as[i] for i in range(len(newshape))] 
        else:
            newshape = self.get_newshape(inputs.shape)
        if self.data_format == 'channels_first':
            return tf.cast(tf.transpose(tf.image.resize_images(tf.transpose(inputs, (0,2,3,1)), newshape, align_corners=self.align_corners, method=self.resize_method), (0,3,1,2)), tf.keras.backend.floatx())
        else:
            return tf.cast(tf.image.resize_images(inputs, newshape, align_corners=self.align_corners, method=self.resize_method), tf.keras.backend.floatx())
        



import tensorflow as tf
import numpy as np

class MergeWithAttention(tf.keras.layers.Add):
    '''
    Multiplies each input with a trainable weight \alpha_i and elementwise adds them.
    '''
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def build(self, input_shape):
        self.attention_weights = self.add_weight(name = 'attention_weights', shape = [len(input_shape), input_shape[0][1]], initializer='uniform', trainable=True)
        super().build(input_shape)
    
    #@tfe.defun
    def _merge_function(self, inputs):
        sm = tf.exp(self.attention_weights)/tf.reduce_sum(tf.exp(self.attention_weights))
        out = tf.einsum('j,ijkl->ijkl',sm[0], inputs[0])
        for i in range(1,len(inputs)):
            out += tf.einsum('j,ijkl->ijkl',sm[i], inputs[i])
        return out
        


import tensorflow as tf
import numpy as np
import itertools
import opt_einsum as oe
import string

from .custom_blocks import ResampledConvolutionBlock, ResnetBlock, AveragePoolingBlock, Scaling
from .Model_With_Integral_Loss import Model_With_Integral_Loss_ABC
from ..layers import MergeWithAttention
from ..layers import Upsample
from ..layers import SpatialPyramidPool
#from ..layers import Scaling

def channels_first_rot_90(image,k=1):
    if len(image.shape) == 4:
        image = tf.transpose(image, (0,2,3,1))
    elif len(image.shape) == 3:
        image = tf.transpose(image, (0,2,1))
        image_was_rank3 = True
    elif len(image.shape) == 2:
        image = tf.expand_dims(image, axis = 2)
        image_was_rank3 = False
    else:
        raise ValueError('image must be a rank 2,3 or 4 Tensor')
    
    image = tf.image.rot90(image, k = k)
    
    if len(image.shape) == 4:
        return tf.transpose(image, (0,3,1,2))
    elif image_was_rank3:
        return tf.transpose(image, (0,2,1))
    else:
        return image[...,0]
    
class Homogeneous_Poisson_NN_Fluidnet(Model_With_Integral_Loss_ABC):
    '''
    Takes a tf.Tensor containing the RHS and grid spacing information, and gives the corresponding solution to the Poisson problem with homogeneous BCs.
    '''
    def __init__(self, pooling_block_number = 6, post_dx_einsum_conv_block_number = 5, initial_kernel_size = 19, final_kernel_size = 3, resize_methods = None, data_format = 'channels_first', use_batchnorm = False, use_deconv_upsample = False, kernel_regularizer = None, bias_regularizer = None, use_scaling = True, scaling_args = {'downsampling_ratio_per_stage': 3, 'stages': 3, 'filters': 4, 'activation': tf.nn.leaky_relu, 'spp_levels': [[2,2],3,5]}, **kwargs):
        '''
        Init arguments:

        data_format: same as keras
        pooling_block_number: integer. controls the number of simultaneous pooling 'threads'. each pooling thread will apply average pooling with pool_size 2^k (k = [1,...,pooling_block_number]), do convoltuions and then upsample to original resolution. choose such that 2^(pooling_block_number) is close as possible to the minimum grid size your inputs will have.
        use_deconv_upsample: boolean. if set to true, pooling blocks will use transposed convolution to upsample. last 2 pooling layers will use bilinear and nearest neighbor unless overridden by resize_methods
        resize_methods: list of tf.image.ResizeMethod members. used to choose resizing algorithms for pooling block upsampling.
        use_batchnorm: boolean. determines if batch norm layers should be used. if set to true, supply __call__ argument training = False when performing inference.
        post_dx_einsum_conv_block_number: integer. no of convolution blocks after the result of the domain info processing is merged.
        initial_kernel_size: integer. initial kernel size of convolutions. most convolutions will have this kernel size.
        final_kernel_size: integer. the kernel size of the final convolutions will progressively move towards this number. keep it small to prevent output artefacting near image edges.
        kernel_regularizer: same as the corresponding tf.keras.layers.Conv2D argument
        bias_regularizer: same as the corresponding tf.keras.layers.Conv2D argument

        **kwargs: used to set self.integral_loss parameters for the Model_With_Integral_Loss_ABC abstract base class
        '''
        super().__init__(**kwargs)
        self.training = True
        self.pooling_block_number = pooling_block_number
        self.data_format = data_format
        self.pooling_block_kernel_sizes = 9 * np.ones((self.pooling_block_number), dtype = np.int32)
        self.pooling_block_kernel_sizes[-2:] = 3
        self.pooling_block_kernel_sizes = list(self.pooling_block_kernel_sizes)
        self.use_batchnorm = use_batchnorm

        final_dense_layer_units = 32
        
        if not resize_methods: #set resizing methods for pooling blocks if using bicubic upsampling
            try:
                self.resize_methods = [tf.compat.v1.image.ResizeMethod.BICUBIC for i in range(self.pooling_block_number-2)] + [tf.compat.v1.image.ResizeMethod.BILINEAR, tf.compat.v1.image.ResizeMethod.NEAREST_NEIGHBOR]
            except:
                self.resize_methods = [tf.image.ResizeMethod.BICUBIC for i in range(self.pooling_block_number-2)] + [tf.image.ResizeMethod.BILINEAR, tf.image.ResizeMethod.NEAREST_NEIGHBOR]
        else:
            self.resize_methods = resize_methods

        self.pooling_block_filters = 32
        
        self.conv_1 = tf.keras.layers.Conv2D(filters = 8, kernel_size = initial_kernel_size, activation=tf.nn.leaky_relu, data_format=data_format, padding='same', kernel_regularizer = kernel_regularizer, bias_regularizer = bias_regularizer)
        #self.batchnorm_1 = tf.keras.layers.BatchNormalization(axis = 1)
        
        self.conv_2 = tf.keras.layers.Conv2D(filters = self.pooling_block_filters, kernel_size = initial_kernel_size, activation=tf.nn.leaky_relu, data_format=data_format, padding='same', kernel_regularizer = kernel_regularizer, bias_regularizer = bias_regularizer)
        #self.batchnorm_2 = tf.keras.layers.BatchNormalization(axis = 1)
        
        self.conv_3 = tf.keras.layers.Conv2D(filters = self.pooling_block_filters, kernel_size = initial_kernel_size, activation=tf.nn.leaky_relu, data_format=data_format, padding='same', kernel_regularizer = kernel_regularizer, bias_regularizer = bias_regularizer)

        if use_deconv_upsample:
            use_deconv_upsample = [True for k in range(self.pooling_block_number-2)] + [False for k in range(2)]
        else:
            use_deconv_upsample = [False for k in range(self.pooling_block_number)]
        self.pooling_blocks = [AveragePoolingBlock(2**(i+1), resize_method = self.resize_methods[i], data_format = data_format, kernel_size = int(self.pooling_block_kernel_sizes[i]), filters = self.pooling_block_filters, activation = tf.nn.leaky_relu, use_resnetblocks = True, use_deconv_upsample = use_deconv_upsample[i], kernel_regularizer = kernel_regularizer, bias_regularizer = bias_regularizer) for i in range(pooling_block_number)]
        
        self.merge = MergeWithAttention()
        
        self.dx_einsum_conv = tf.keras.layers.Conv2D(filters = final_dense_layer_units, kernel_size = initial_kernel_size, activation=tf.nn.leaky_relu, data_format=data_format, padding='same', kernel_regularizer = kernel_regularizer, bias_regularizer = bias_regularizer)
        self.dx_einsum_resnet = ResnetBlock(filters = final_dense_layer_units, kernel_size = initial_kernel_size, activation=tf.nn.leaky_relu, data_format=data_format, kernel_regularizer = kernel_regularizer, bias_regularizer = bias_regularizer)

        self.post_dx_einsum_conv_blocks = []
        for k in range(post_dx_einsum_conv_block_number):
            ksize = initial_kernel_size + (k*(final_kernel_size - initial_kernel_size)//(post_dx_einsum_conv_block_number-1))
            filters = final_dense_layer_units-((k*final_dense_layer_units)//(post_dx_einsum_conv_block_number))
            block = []
            block.append(tf.keras.layers.Conv2D(filters = filters, kernel_size = ksize, activation=tf.nn.leaky_relu, data_format=data_format, padding='same', kernel_regularizer = kernel_regularizer, bias_regularizer = bias_regularizer))
            block.append(ResnetBlock(filters = filters, kernel_size = ksize, activation=tf.nn.leaky_relu, data_format=data_format, kernel_regularizer = kernel_regularizer, bias_regularizer = bias_regularizer))
            self.post_dx_einsum_conv_blocks.append(block)

        if use_batchnorm:
            self.batchnorm_layers = []
            if self.data_format == 'channels_first':
                self.batchnorm_last = tf.keras.layers.BatchNormalization(axis = 1)
            else:
                self.batchnorm_last = tf.keras.layers.BatchNormalization(axis = -1)
            for k in range(post_dx_einsum_conv_block_number):
                if self.data_format == 'channels_first':
                    self.batchnorm_layers.append(tf.keras.layers.BatchNormalization(axis = 1))
                else:
                    self.batchnorm_layers.append(tf.keras.layers.BatchNormalization(axis = -1))

        self.conv_last = tf.keras.layers.Conv2D(filters = 1, kernel_size = final_kernel_size, activation='linear', data_format=data_format, padding='same', kernel_regularizer = kernel_regularizer, bias_regularizer = bias_regularizer)
        self.resnet_last = ResnetBlock(filters = 1, kernel_size = final_kernel_size, activation='linear', data_format=data_format, kernel_regularizer = kernel_regularizer, bias_regularizer = bias_regularizer)
        
        self.dx_dense_0 = tf.keras.layers.Dense(100, activation = tf.nn.leaky_relu)
        self.dx_dense_1 = tf.keras.layers.Dense(100, activation = tf.nn.leaky_relu)
        self.dx_dense_2 = tf.keras.layers.Dense(final_dense_layer_units, activation = 'linear')

        self.use_scaling = use_scaling
        if use_scaling:
            self.scaling = Scaling(data_format = self.data_format, kernel_size = final_kernel_size, kernel_regularizer = kernel_regularizer, bias_regularizer = bias_regularizer, **scaling_args)
        
    def call(self, inp):
        '''
        input arguments:
        
        inp[0]: tf.Tensor of shape (batch_size, 1, nx, ny) or (batch_size, nx, ny, 1) based on self.data_format. this is the RHSes.
        inp[1]: tf.Tensor of shape (batch_size, 1). this must be the grid spacing information.
        '''
        self.dx = inp[1]
        if self.data_format == 'channels_first':#compute domain info
            domain_info = tf.einsum('ij,j->ij',tf.tile(inp[1], [1,3]), tf.stack([tf.constant(1.0, dtype = tf.keras.backend.floatx()), tf.cast(tf.constant(inp[0].shape[2]), tf.keras.backend.floatx()), tf.cast(tf.constant(inp[0].shape[3]), tf.keras.backend.floatx())], axis = 0))
        else:
            domain_info = tf.einsum('ij,j->ij',tf.tile(inp[1], [1,3]), tf.stack([tf.constant(1.0, dtype = tf.keras.backend.floatx()), tf.cast(tf.constant(inp[0].shape[1]), tf.keras.backend.floatx()), tf.cast(tf.constant(inp[0].shape[2]), tf.keras.backend.floatx())], axis = 0))

        out = self.conv_1(inp[0]) #initial convolutions
        out = self.conv_2(out)
        
        out = self.merge([self.conv_3(out)] + [pb(out) for pb in self.pooling_blocks])#pooling 'threads'

        out = self.dx_einsum_conv(out)
        out = self.dx_einsum_resnet(out)
        
        out = tf.einsum('ijkl, ij -> ijkl',out, self.dx_dense_2(self.dx_dense_1(self.dx_dense_0(domain_info)))) #incorporate domain info

        for block_num, block in enumerate(self.post_dx_einsum_conv_blocks): #perform additional convolutions
            for layer in block:
                out = layer(out)
            if self.use_batchnorm:
                out = self.batchnorm_layers[block_num](out, training = self.training)

        if self.use_batchnorm:
            out = self.batchnorm_last(out, training = self.training)
        out = self.conv_last(out)
        out = self.resnet_last(out)

        if self.use_scaling:
            out = self.scaling([out, inp[0]])

        if self.training == False:
            out = tf.einsum('i...,i->i...', out, tf.reduce_prod(domain_info[:,1:], axis = 1))
            
        return out

    def __call__(self, inp, training = True):#overload __call__ to allow freezing batch norm parameters
        self.training = training
        return super().__call__(inp)
    

class Homogeneous_Poisson_NN_Fourier(Model_With_Integral_Loss_ABC): #variant to include dx info
    def __init__(self, filters = [8,16,32,64,16,8], kernel_sizes = [5,5,5,5,5,5], pool_after_kth_resnet_block = None, pool_size = 2, nmodes = (32,32), pyramid_pooling_params = {'levels': [[3,3],6,9,12], 'pooling_type': 'AVG'}, data_format = 'channels_first', domain_info_dense_layer_units = [100,100,16], final_dense_layer_units = [250, 500], kernel_regularizer = None, bias_regularizer = None, **kwargs):
        super().__init__(**kwargs)
        self.data_format = data_format
        if self.data_format == 'channels_first':
            self.space_dims_start = 2
        else:
            self.space_dims_start = 1
        self.space_dims_end = self.space_dims_start + len(nmodes)
        self.resnet_blocks = []
        self.nmodes = nmodes
        self.total_mode_number = int(tf.reduce_prod(nmodes))
        
        for i in range(len(kernel_sizes)):
            self.resnet_blocks.append(tf.keras.layers.Conv2D(filters = filters[i], kernel_size = kernel_sizes[i], data_format = self.data_format, activation = tf.nn.leaky_relu, padding = 'same', kernel_regularizer = kernel_regularizer, bias_regularizer = bias_regularizer))
            self.resnet_blocks.append(ResnetBlock(filters = filters[i], kernel_size = kernel_sizes[i], data_format = self.data_format, activation = tf.nn.leaky_relu, kernel_regularizer = kernel_regularizer, bias_regularizer = bias_regularizer))
            if (pool_after_kth_resnet_block is not None) and (i % pool_after_kth_resnet_block == 0):
                self.resnet_blocks.append(tf.keras.layers.MaxPool2D(pool_size = 2, padding = 'same', data_format = self.data_format))
        self.spp = SpatialPyramidPool(data_format = self.data_format, **pyramid_pooling_params)

        self.domain_info_dense_layers = []
        for units in domain_info_dense_layer_units:
            self.domain_info_dense_layers.append(tf.keras.layers.Dense(units, activation = tf.nn.leaky_relu, kernel_regularizer = kernel_regularizer, bias_regularizer = bias_regularizer))

        self.final_dense_layers = []
        for units in final_dense_layer_units:
            self.final_dense_layers.append(tf.keras.layers.Dense(units, activation = tf.nn.leaky_relu, kernel_regularizer = kernel_regularizer, bias_regularizer = bias_regularizer))
        self.final_dense_layers.append(tf.keras.layers.Dense(tf.reduce_prod(nmodes), activation = lambda x: 4*tf.nn.tanh(x)/np.pi**2))

        self.mplus1_pi_squared = tf.cast(tf.stack(tf.meshgrid(*[((np.arange(mode_number)+1)*np.pi)**2 for mode_number in self.nmodes], indexing = 'ij'), axis = 0), tf.keras.backend.floatx())
        # self.dx_dense_0 = tf.keras.layers.Dense(100, activation = tf.nn.leaky_relu)
        # self.dx_dense_1 = tf.keras.layers.Dense(100, activation = tf.nn.leaky_relu)
        # self.dx_dense_2 = tf.keras.layers.Dense(16, activation = tf.nn.leaky_relu)
        
        # self.fourier_dense_0 = tf.keras.layers.Dense(250, activation = tf.nn.leaky_relu)
        # self.fourier_dense_1 = tf.keras.layers.Dense(500, activation = tf.nn.leaky_relu)
        # self.fourier_dense_2 = tf.keras.layers.Dense(tf.reduce_prod(nmodes), activation = lambda x: 4*tf.nn.tanh(x)/np.pi**2)

        # self.scaling = Scaling()

        
        # m,n = tf.meshgrid(np.arange(self.nmodes[0])+1,np.arange(self.nmodes[1])+1)
        # self.m = tf.expand_dims((tf.cast(m, tf.keras.backend.floatx())**2) * (tf.cast(np.pi, tf.keras.backend.floatx())**2), axis = 0)
        # self.n = tf.expand_dims((tf.cast(n, tf.keras.backend.floatx())**2) * (tf.cast(np.pi, tf.keras.backend.floatx())**2), axis = 0)

    def call(self,inp):
        inpshape = tf.keras.backend.shape(inp[0])
        self.dx = inp[1]
        domain_info = tf.einsum('ij,j->ij',tf.tile(inp[1], [1,3]), tf.keras.backend.stack([tf.constant(1.0, dtype = tf.keras.backend.floatx())] + [tf.cast(inpshape[k], tf.keras.backend.floatx()) for k in range(self.space_dims_start, self.space_dims_end)], axis = 0))
        
        grids = tf.meshgrid(*[tf.linspace(0.0,1.0,tf.keras.backend.cast(inpshape[k], tf.int32)) for k in range(self.space_dims_start, self.space_dims_end)], indexing = 'ij')
        for grid in grids:
            grid = tf.cast(grid, tf.keras.backend.floatx())
            
        sine_values = tf.Variable(lambda: tf.tile(tf.expand_dims(tf.keras.backend.zeros(tf.shape(grids[0]), dtype = tf.keras.backend.floatx()), axis = 0), [self.total_mode_number] + [1 for dim in grids[0].shape]))
        for m in range(self.total_mode_number):
            indices = np.zeros((len(self.nmodes),), dtype = np.int64)
            indices[-1] = m % self.nmodes[-1]
            for k in range(len(self.nmodes)-1):
                indices[k] = int(np.floor(m/np.prod(self.nmodes[k+1:]))) % self.nmodes[k]
            sine_val = tf.Variable(lambda: tf.sin((indices[0]+1)*np.pi*grids[0]))
            for k in range(1, len(indices)):
                sine_val.assign(sine_val * tf.sin((indices[k]+1)*np.pi*grids[k]))
            sine_values[m].assign(sine_val)

        sine_values = tf.keras.backend.reshape(sine_values, tf.concat([list(self.nmodes) , tf.shape(grids[0])], axis = 0))

        dx_res = self.domain_info_dense_layers[0](domain_info)
        for layer in self.domain_info_dense_layers[1:]:
            dx_res = layer(dx_res)

        amplitudes = self.resnet_blocks[0](inp[0])
        for layer in self.resnet_blocks[1:]:
            amplitudes = layer(amplitudes)
        amplitudes = tf.concat([self.spp(amplitudes),dx_res], axis = 1)

        for layer in self.final_dense_layers:
            amplitudes = layer(amplitudes)
            
        amplitudes = tf.keras.backend.reshape(amplitudes, [-1] + list(self.nmodes))
        amplitudes = -amplitudes * (2**(len(inpshape)-2)) / tf.einsum('bi,i...->b...', 1/(domain_info[:,1:])**2,self.mplus1_pi_squared)

        contr_expr = 'z' + string.ascii_lowercase[:len(amplitudes.shape)-1] + ',' + string.ascii_lowercase[:len(amplitudes.shape)-1] + '...->z...'
        out = tf.einsum(contr_expr, amplitudes, sine_values)

        if self.data_format == 'channels_first':
            return tf.expand_dims(out, axis = 1)
        else:
            return tf.expand_dims(out, axis = -1)

################# TEST SECTION #####################
from ..dataset.generators.numerical import set_max_magnitude_in_batch as smmib
class Scaling_Test(tf.keras.models.Model):
    def __init__(self, stages = 2, downsampling_ratio_per_stage = 2, data_format = 'channels_first', padding = 'same', spp_levels = [[2,2], 3, 5], **convargs):
        super().__init__()
        self.spp = SpatialPyramidPool(spp_levels, data_format = data_format, pooling_type = 'MAX')
        self.stages = []
        self.data_format = data_format
        for k in range(stages):
            self.stages.append(tf.keras.layers.Conv2D(data_format = data_format, padding = padding, **convargs))
            self.stages.append(tf.keras.layers.AveragePooling2D(pool_size = downsampling_ratio_per_stage, padding = padding, data_format = data_format))

        self.dense_0 = tf.keras.layers.Dense(100, activation = tf.nn.leaky_relu)
        self.dense_1 = tf.keras.layers.Dense(25, activation = tf.nn.leaky_relu)
        self.dense_2 = tf.keras.layers.Dense(1, activation = 'linear')

    def call(self, inp):
        #input format: tf.Tensor of shape [batch_size, 2, nx, ny] where [:,0,...] is the input to conditionally scale, or a list of 2 tf.Tensors such that [input_to_scale, 2nd input]
        domain_info = inp[2]
        if self.data_format == 'channels_first' and isinstance(inp, list):
            inp = tf.concat(inp[:2], axis = 1)
            inp_to_scale = tf.expand_dims(inp[:,0,...], axis = 1)
        else:
            inp = tf.concat(inp[:2], axis = -1)
            inp_to_scale = tf.expand_dims(inp[:,...,0], axis = -1)
        
        out = self.stages[0](inp)
        for k in range(1, len(self.stages)):
            out = self.stages[k](out)
        out = tf.concat([self.spp(out), domain_info], axis = 1)
        out = self.dense_2(self.dense_1(self.dense_0(out)))
        return out
    
class Homogeneous_Poisson_NN_Fluidnet_Test(Model_With_Integral_Loss_ABC):
    '''
    Takes a tf.Tensor containing the RHS and grid spacing information, and gives the corresponding solution to the Poisson problem with homogeneous BCs.
    '''
    def __init__(self, pooling_block_number = 6, post_dx_einsum_conv_block_number = 5, initial_kernel_size = 19, final_kernel_size = 3, resize_methods = None, data_format = 'channels_first', use_batchnorm = False, use_deconv_upsample = False, kernel_regularizer = None, bias_regularizer = None, use_scaling = True, output_max_magnitude = 1.0, scaling_args = {'downsampling_ratio_per_stage': 3, 'stages': 3, 'filters': 4, 'activation': tf.nn.leaky_relu, 'spp_levels': [[2,2],3,5]}, **kwargs):
        '''
        Init arguments:

        data_format: same as keras
        pooling_block_number: integer. controls the number of simultaneous pooling 'threads'. each pooling thread will apply average pooling with pool_size 2^k (k = [1,...,pooling_block_number]), do convoltuions and then upsample to original resolution. choose such that 2^(pooling_block_number) is close as possible to the minimum grid size your inputs will have.
        use_deconv_upsample: boolean. if set to true, pooling blocks will use transposed convolution to upsample. last 2 pooling layers will use bilinear and nearest neighbor unless overridden by resize_methods
        resize_methods: list of tf.image.ResizeMethod members. used to choose resizing algorithms for pooling block upsampling.
        use_batchnorm: boolean. determines if batch norm layers should be used. if set to true, supply __call__ argument training = False when performing inference.
        post_dx_einsum_conv_block_number: integer. no of convolution blocks after the result of the domain info processing is merged.
        initial_kernel_size: integer. initial kernel size of convolutions. most convolutions will have this kernel size.
        final_kernel_size: integer. the kernel size of the final convolutions will progressively move towards this number. keep it small to prevent output artefacting near image edges.
        kernel_regularizer: same as the corresponding tf.keras.layers.Conv2D argument
        bias_regularizer: same as the corresponding tf.keras.layers.Conv2D argument

        **kwargs: used to set self.integral_loss parameters for the Model_With_Integral_Loss_ABC abstract base class
        '''
        super().__init__(**kwargs)
        self.training = True
        self.pooling_block_number = pooling_block_number
        self.data_format = data_format
        self.pooling_block_kernel_sizes = 9 * np.ones((self.pooling_block_number), dtype = np.int32)
        self.pooling_block_kernel_sizes[-2:] = 3
        self.pooling_block_kernel_sizes = list(self.pooling_block_kernel_sizes)
        self.use_batchnorm = use_batchnorm
        self.output_max_magnitude = output_max_magnitude

        final_dense_layer_units = 32
        
        if not resize_methods: #set resizing methods for pooling blocks if using bicubic upsampling
            try:
                self.resize_methods = [tf.compat.v1.image.ResizeMethod.BICUBIC for i in range(self.pooling_block_number-2)] + [tf.compat.v1.image.ResizeMethod.BILINEAR, tf.compat.v1.image.ResizeMethod.NEAREST_NEIGHBOR]
            except:
                self.resize_methods = [tf.image.ResizeMethod.BICUBIC for i in range(self.pooling_block_number-2)] + [tf.image.ResizeMethod.BILINEAR, tf.image.ResizeMethod.NEAREST_NEIGHBOR]
        else:
            self.resize_methods = resize_methods

        self.pooling_block_filters = 24
        
        self.conv_1 = tf.keras.layers.Conv2D(filters = 8, kernel_size = initial_kernel_size, activation=tf.nn.leaky_relu, data_format=data_format, padding='same', kernel_regularizer = kernel_regularizer, bias_regularizer = bias_regularizer)
        #self.batchnorm_1 = tf.keras.layers.BatchNormalization(axis = 1)
        
        self.conv_2 = tf.keras.layers.Conv2D(filters = self.pooling_block_filters, kernel_size = initial_kernel_size, activation=tf.nn.leaky_relu, data_format=data_format, padding='same', kernel_regularizer = kernel_regularizer, bias_regularizer = bias_regularizer)
        #self.batchnorm_2 = tf.keras.layers.BatchNormalization(axis = 1)
        
        self.conv_3 = tf.keras.layers.Conv2D(filters = self.pooling_block_filters, kernel_size = initial_kernel_size, activation=tf.nn.leaky_relu, data_format=data_format, padding='same', kernel_regularizer = kernel_regularizer, bias_regularizer = bias_regularizer)

        if use_deconv_upsample:
            use_deconv_upsample = [True for k in range(self.pooling_block_number-2)] + [False for k in range(2)]
        else:
            use_deconv_upsample = [False for k in range(self.pooling_block_number)]
        self.pooling_blocks = [AveragePoolingBlock(2**(i+1), resize_method = self.resize_methods[i], data_format = data_format, kernel_size = int(self.pooling_block_kernel_sizes[i]), filters = self.pooling_block_filters, activation = tf.nn.leaky_relu, use_resnetblocks = True, use_deconv_upsample = use_deconv_upsample[i], kernel_regularizer = kernel_regularizer, bias_regularizer = bias_regularizer) for i in range(pooling_block_number)]
        
        self.merge = MergeWithAttention()
        
        self.dx_einsum_conv = tf.keras.layers.Conv2D(filters = final_dense_layer_units, kernel_size = initial_kernel_size, activation=tf.nn.leaky_relu, data_format=data_format, padding='same', kernel_regularizer = kernel_regularizer, bias_regularizer = bias_regularizer)
        self.dx_einsum_resnet = ResnetBlock(filters = final_dense_layer_units, kernel_size = initial_kernel_size, activation=tf.nn.leaky_relu, data_format=data_format, kernel_regularizer = kernel_regularizer, bias_regularizer = bias_regularizer)

        self.post_dx_einsum_conv_blocks = []
        for k in range(post_dx_einsum_conv_block_number):
            ksize = initial_kernel_size + (k*(final_kernel_size - initial_kernel_size)//(post_dx_einsum_conv_block_number-1))
            filters = final_dense_layer_units-((k*final_dense_layer_units)//(post_dx_einsum_conv_block_number))
            block = []
            block.append(tf.keras.layers.Conv2D(filters = filters, kernel_size = ksize, activation=tf.nn.leaky_relu, data_format=data_format, padding='same', kernel_regularizer = kernel_regularizer, bias_regularizer = bias_regularizer))
            block.append(ResnetBlock(filters = filters, kernel_size = ksize, activation=tf.nn.leaky_relu, data_format=data_format, kernel_regularizer = kernel_regularizer, bias_regularizer = bias_regularizer))
            self.post_dx_einsum_conv_blocks.append(block)

        self.residual_pooling_connection_blocks = []
        for k in range(3):
            block = []
            block.append(tf.keras.layers.Conv2D(filters = filters, kernel_size = final_kernel_size, activation = tf.nn.leaky_relu, data_format = data_format, padding = 'same', kernel_regularizer = kernel_regularizer, bias_regularizer = bias_regularizer))
            block.append(ResnetBlock(filters = filters, kernel_size = final_kernel_size, activation = tf.nn.leaky_relu, data_format = data_format, kernel_regularizer = kernel_regularizer, bias_regularizer = bias_regularizer))
            self.residual_pooling_connection_blocks.append(block)

        if use_batchnorm:
            self.batchnorm_layers = []
            if self.data_format == 'channels_first':
                self.batchnorm_last = tf.keras.layers.BatchNormalization(axis = 1)
            else:
                self.batchnorm_last = tf.keras.layers.BatchNormalization(axis = -1)
            for k in range(post_dx_einsum_conv_block_number):
                if self.data_format == 'channels_first':
                    self.batchnorm_layers.append(tf.keras.layers.BatchNormalization(axis = 1))
                else:
                    self.batchnorm_layers.append(tf.keras.layers.BatchNormalization(axis = -1))

        self.conv_penultimate = tf.keras.layers.Conv2D(filters = filters, kernel_size = final_kernel_size, activation=tf.nn.leaky_relu, data_format=data_format, padding='same', kernel_regularizer = kernel_regularizer, bias_regularizer = bias_regularizer)
        self.resnet_penultimate = ResnetBlock(filters = filters, kernel_size = final_kernel_size, activation=tf.nn.leaky_relu, data_format=data_format, kernel_regularizer = kernel_regularizer, bias_regularizer = bias_regularizer)

        self.merge_final = MergeWithAttention()
        self.conv_last = tf.keras.layers.Conv2D(filters = 1, kernel_size = final_kernel_size, activation=tf.nn.tanh, data_format=data_format, padding='same', kernel_regularizer = kernel_regularizer, bias_regularizer = bias_regularizer)
        self.resnet_last = ResnetBlock(filters = 1, kernel_size = final_kernel_size, activation=tf.nn.tanh, data_format=data_format, kernel_regularizer = kernel_regularizer, bias_regularizer = bias_regularizer)
        self.output_denoiser = AveragePoolingBlock(pool_size = 4, data_format=data_format, use_resnetblocks = True, use_deconv_upsample = False, kernel_size = final_kernel_size, filters = 1, activation=tf.nn.tanh, kernel_regularizer = kernel_regularizer, bias_regularizer = bias_regularizer) #13.09 set activations to tanh
        
        self.dx_dense_0 = tf.keras.layers.Dense(100, activation = tf.nn.leaky_relu)
        self.dx_dense_1 = tf.keras.layers.Dense(100, activation = tf.nn.leaky_relu)
        self.dx_dense_2 = tf.keras.layers.Dense(final_dense_layer_units, activation = 'linear', kernel_regularizer = kernel_regularizer, bias_regularizer = bias_regularizer)


        self.use_scaling = use_scaling
        if use_scaling:
            #self.scaling = Scaling(data_format = self.data_format, kernel_size = final_kernel_size, kernel_regularizer = kernel_regularizer, bias_regularizer = bias_regularizer, **scaling_args)

            self.dx_dense_scaler_0 = tf.keras.layers.Dense(30, activation = tf.nn.leaky_relu)
            self.dx_dense_scaler_1 = tf.keras.layers.Dense(30, activation = tf.nn.leaky_relu)
            self.dx_dense_scaler_2 = tf.keras.layers.Dense(1, activation = 'linear')
        
    def call(self, inp):
        '''
        input arguments:
        
        inp[0]: tf.Tensor of shape (batch_size, 1, nx, ny) or (batch_size, nx, ny, 1) based on self.data_format. this is the RHSes.
        inp[1]: tf.Tensor of shape (batch_size, 1). this must be the grid spacing information.
        '''
        self.dx = inp[1]
        if self.data_format == 'channels_first':#compute domain info
            domain_info = tf.einsum('ij,j->ij',tf.tile(inp[1], [1,3]), tf.stack([tf.constant(1.0, dtype = tf.keras.backend.floatx()), tf.cast(tf.constant(inp[0].shape[2]), tf.keras.backend.floatx()), tf.cast(tf.constant(inp[0].shape[3]), tf.keras.backend.floatx())], axis = 0))
        else:
            domain_info = tf.einsum('ij,j->ij',tf.tile(inp[1], [1,3]), tf.stack([tf.constant(1.0, dtype = tf.keras.backend.floatx()), tf.cast(tf.constant(inp[0].shape[1]), tf.keras.backend.floatx()), tf.cast(tf.constant(inp[0].shape[2]), tf.keras.backend.floatx())], axis = 0))

        #inv_dom_size = tf.reduce_prod(1/domain_info[:,1:], axis = 1, keepdims = True)
        #domain_info = tf.concat([domain_info, inv_dom_size], axis = 1)
        
        out = self.conv_1(inp[0]) #initial convolutions
        out = self.conv_2(out)

        pooling_block_result = [pb(out) for pb in self.pooling_blocks]
        out = self.merge([self.conv_3(out)]+pooling_block_result)#pooling 'threads'

        out = self.dx_einsum_conv(out)
        out = self.dx_einsum_resnet(out)

        dx_res = self.dx_dense_2(self.dx_dense_1(self.dx_dense_0(domain_info)))
        
        out = tf.einsum('ijkl, ij -> ijkl',out, dx_res) #incorporate domain info

        for block_num, block in enumerate(self.post_dx_einsum_conv_blocks): #perform additional convolutions
            for layer in block:
                out = layer(out)
            if self.use_batchnorm:
                out = self.batchnorm_layers[block_num](out, training = self.training)

        if self.use_batchnorm:
            out = self.batchnorm_last(out, training = self.training)
        out = self.resnet_penultimate(self.conv_penultimate(out))

        rpcb_res = []
        for blk_num, blk in enumerate(self.residual_pooling_connection_blocks):
            tmp = blk[0](pooling_block_result[blk_num])
            for lyr in blk[1:]:
                tmp = lyr(tmp)
            rpcb_res.append(tmp)

        out = self.merge_final([out]+rpcb_res)
        out = self.resnet_last(self.conv_last(out))
        #out += sum(rpcb_res)
        out = self.output_denoiser(out)
            
        if self.use_scaling:
            #out = self.scaling([out, inp[0], dx_res])
            try:
                scale = self.dx_dense_scaler_2(self.dx_dense_scaler_1(self.dx_dense_scaler_0(domain_info)))
                out = tf.einsum('i,i...->i...',tf.squeeze(scale),out)
            except:
                pass

        if self.training == False:
            out = tf.einsum('i...,i->i...', out, tf.reduce_prod(domain_info[:,1:], axis = 1))

        if self.output_max_magnitude is not None:
            out = smmib(out, 1.0)#13.09 added smmib

        return out

    def __call__(self, inp, training = True):#overload __call__ to allow freezing batch norm parameters
        self.training = training
        return super().__call__(inp)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Jan 10 13:36:21 2019

@author: ali
"""
import numpy as np
import scipy
#from inspect import signature
import tensorflow as tf
#import pdb

class Boundary1D:
    '''
        A class that implements a 1D curved boundary defined via piecewise interpolation of provided coordinates
        
        Init parameters:
            boundary_type               : 'Dirichlet' (u = g on the boundary), 'von Neumann' (du/dn = g on the boundary) or 'Robin' (du/dn + a*u = g on the boundary)
            coordinates                 : a list or numpy array containing the coordinates of the boundary supplied in order
            error_type                  : Currently Lp spaces are supported
            RHS_function                : g. Supply a function taking a single argument. If it's a parametric function, set boundary_rhs_is_parametric to True. If not, (i.e. if the function takes x and y coords) it should  still take a single argument but the supplied argument should contain x = arg[0] and y = arg[1]
            robin_bc_alpha              : value of a used in the Robin BC
            interpolation_order         : order of spline interpolation to be used between supplied coords.
            orientation                 : clockwise boundaries will rotate normal vector 90 degrees CCW of the tangent and vice versa
            boundary_rhs_is_parametric  : determines if a parametric function g(t) or a 'regular' function g(x,y) is specified as the rhs of the boundary
    '''
    
    def __init__(self, boundary_type, coordinates, error_type = 'L2', RHS_function = lambda x: 0, label = None, robin_bc_alpha = None, interpolation_order = 1, orientation = 'counterclockwise', derivative_regularizer_coefficient = 0.0, boundary_rhs_is_parametric = False, n_threads = 8, dtype = 'float64'):
        
        self.dtype = dtype
        self.error_type = error_type
        self.orientation = orientation
        self.boundary_type = boundary_type
        self.coordinates = np.array(coordinates)
        self.label = label
        self.RHS_function = RHS_function
        if robin_bc_alpha:
            self.robin_bc_alpha = robin_bc_alpha
        
        self.t = self.get_arc_length(self.coordinates)
        self.x_interpolant = scipy.interpolate.InterpolatedUnivariateSpline(self.t,self.coordinates[:,0], k = interpolation_order)
        self.y_interpolant = scipy.interpolate.InterpolatedUnivariateSpline(self.t,self.coordinates[:,1], k = interpolation_order)
        self.dxdt = self.x_interpolant.derivative(n = 1)
        self.dydt = self.y_interpolant.derivative(n = 1)
        self.derivative_regularizer_coefficient = derivative_regularizer_coefficient
        self.boundary_rhs_is_parametric = boundary_rhs_is_parametric        
        self.n_threads = 8
        
        if self.orientation == 'counterclockwise':
            self.tangent_to_normal_rotation_matrix = np.array([[0,1],[-1,0]])
        elif self.orientation == 'clockwise':
            self.tangent_to_normal_rotation_matrix = np.array([[0,-1],[1,0]])
    
    #@tf.contrib.eager.defun #wrapper function for the RHS function to permit parallelization
    def RHS_evaluate(self, x):
        return tf.map_fn(self.RHS_function, x, dtype = self.dtype)
            
    def get_tangent_vectors(self,t_values, return_unit_vectors = False): #get tangent vectors to the boundary at the specified pts
        tmp = np.array([self.dxdt(t_values), self.dydt(t_values)])
        if return_unit_vectors:
            return np.divide(tmp, np.linalg.norm(tmp, axis = 0))
        else:
            return tmp
        
    def get_normal_vectors(self,t_values, return_unit_vectors = False): #get normal vectors to the boundary at the specified pts
        return np.einsum('ij,jk->ik', self.tangent_to_normal_rotation_matrix, self.get_tangent_vectors(t_values, return_unit_vectors = return_unit_vectors))    
    
    @staticmethod
    def get_arc_length(coordinates): #approximates arc length along the boundary for a curve fit
        return np.cumsum(np.insert(np.sqrt(np.sum(np.square(np.subtract(np.array(coordinates[1:]),np.array(coordinates[:-1]))), axis = 1)),0,0.0))
    
    def LHS_evaluate(self, model, t_values): #evaluates the LHS at the specified pts., LHS has to be a keras model or otherwise return a mapping (x,y)->u from a method named 'predict_on_batch' when given the coordinates desired as a 2d tensorflow tensor
         x = self.x_interpolant(t_values)
         y = self.y_interpolant(t_values)
         
         self.last_used_pts = tf.constant(np.array([x,y]).T, dtype = self.dtype)
         
         if self.boundary_type == 'Dirichlet':
             if self.derivative_regularizer_coefficient == 0.0:
                 return model.predict_on_batch(self.last_used_pts), None
             else: #optionally calculates du/dt if needed
                 with tf.GradientTape() as tape:
                     tape.watch(self.last_used_pts)
                     y = model.predict_on_batch(self.last_used_pts)
                 grad = tape.gradient(y, self.last_used_pts)
                 dydt = tf.einsum('ij,ij->i',grad,tf.constant(self.get_tangent_vectors(t_values, return_unit_vectors = True).T,dtype = self.dtype))
                 return y, dydt
                 
         elif self.boundary_type == 'von Neumann' or self.boundary_type == 'Robin':
             
             with tf.GradientTape(persistent=True) as tape:
                 tape.watch(self.last_used_pts)
                 y = model.predict_on_batch(self.last_used_pts)
             grad = tape.gradient(y, self.last_used_pts)
             normal_gradients = tf.einsum('ij,ij->i',grad,tf.constant(self.get_normal_vectors(t_values, return_unit_vectors = True).T,dtype = self.dtype))
             if self.derivative_regularizer_coefficient == 0.0:
                 dydt = None
             else:
                 dydt = tf.einsum('ij,ij->i',grad,tf.constant(self.get_tangent_vectors(t_values, return_unit_vectors = True).T,dtype = self.dtype))
             
             if self.boundary_type == 'von Neumann':
                 return normal_gradients, dydt
             elif self.boundary_type == 'Robin':
                 return normal_gradients + self.robin_bc_alpha * y, dydt
    
    def evaluate_error(self, model, n_quadpts = 5): #evaluates error
        #Evaluates the error between the values provided by the model and the RHS using Gauss-Legendre quadrature
        quadrature_t, quadrature_w = tuple([np.polynomial.legendre.leggauss(n_quadpts)[i].astype(self.dtype) for i in range(2)]) #get quadrature pts
        c = np.array(0.5*(self.t[-1] - self.t[0]),dtype=self.dtype)#helper variables
        d = np.array(0.5*(self.t[-1] + self.t[0]),dtype=self.dtype)
        quadpts = tf.constant(c*quadrature_t+d, dtype = self.dtype) #quadrature points on boundary
        #pdb.set_trace()
        LHS, LHSderivatives = self.LHS_evaluate(model, quadpts) #LHS values
        
        #RHS values
        if self.boundary_rhs_is_parametric:
            if self.derivative_regularizer_coefficient == 0.0:
                RHS = self.RHS_evaluate(quadpts)
            else:
                with tf.GradientTape() as tape:
                    tape.watch(quadpts)
                    RHS = self.RHS_evaluate(quadpts)
                RHSderivatives = tape.gradient(RHS, quadpts)
        else:
            if self.derivative_regularizer_coefficient == 0.0:
                RHS = self.RHS_evaluate(quadpts)
            else:
                with tf.GradientTape() as tape:
                    tape.watch(self.last_used_pts)
                    RHS = self.RHS_evaluate(quadpts)
                RHSderivatives = tf.einsum('ij,ij->i', tape.gradient(RHS, self.last_used_pts), tf.constant(self.get_tangent_vectors(quadpts, return_unit_vectors = True).T,dtype = self.dtype))
        
        #Calculate Lp norm and return
        if len(self.error_type) == 2 and self.error_type[0] == 'L':
            Lp_norm_order = int(self.error_type[1])
            error_function = tf.pow(tf.reduce_sum(tf.multiply(c*quadrature_w,tf.pow(tf.reshape(LHS, [tf.reduce_prod(LHS.shape)])-RHS,Lp_norm_order))),1/Lp_norm_order)
            if LHSderivatives != None:
                error_derivatives = tf.pow(tf.reduce_sum(tf.multiply(c*quadrature_w,tf.pow(LHSderivatives-RHSderivatives,Lp_norm_order))),1/Lp_norm_order)
            else:
                error_derivatives = 0.0
            return error_function + self.derivative_regularizer_coefficient * error_derivatives
                



import tensorflow as tf
import numpy as np

class SpatialPyramidPool(tf.keras.layers.Layer):
    def __init__(self, levels, data_format = 'channels_first', pooling_type = 'AVG' ,**kwargs):
        super().__init__(**kwargs)
        self.maxdims = 3
        self.data_format = data_format
        self.pooling_type = pooling_type
        self.levels = levels
        #self.data_format = data_format
        self.ndims = None
        #check if all SPP levels have the same number of dimensions, fix dimensions along which only ints were given
        level_dimensionalities = []
        for level in self.levels:
            try:
                level_dimensionalities.append(len(level))
            except:
                level_dimensionalities.append(0)
        level_dimensionalities = np.array(level_dimensionalities)
        if np.all(level_dimensionalities == 0):
            print('Spatial Pyramid Pooling Layer Warning: Input dimensionality can not be inferred. This can negatively impact performance.')
            return    
        elif np.any(level_dimensionalities[level_dimensionalities!=0] != level_dimensionalities[level_dimensionalities !=0][0]):
            raise(ValueError('Dimension number mismatch in the provided level output shapes'))
        else:
            self.ndims = level_dimensionalities[level_dimensionalities != 0][0]
            for i in range(len(self.levels)):
                if isinstance(self.levels[i], int):
                    s = int(self.levels[i])
                    self.levels[i] = [s for i in range(self.ndims)]
            if self.ndims > self.maxdims:
                raise(ValueError('ndims must be between 1 and ' + str(self.maxdims)))
            elif self.data_format == 'channels_first':
                self.data_format = 'NC' + 'DHW'[(self.maxdims-self.ndims):]
            else:
                self.data_format = 'N' + 'DHW'[(self.maxdims-self.ndims):] + 'C'
    #@tf.function
    def call(self, inp):
        if self.ndims == None: #build the pooling levels based on input shape if not already determined
            ndims = len(inp.shape) - 2
            levels = []
            for i in range(len(self.levels)):
                s = int(self.levels[i])
                levels.append([s for i in range(ndims)])
            if ndims > self.maxdims:
                raise(ValueError('ndims must be between 1 and ' + str(self.maxdims)))
            elif self.data_format == 'channels_first':
                data_format = 'NC' + 'DHW'[(self.maxdims-ndims):]
            else:
                data_format = 'N' + 'DHW'[(self.maxdims-ndims):] + 'C'
        else:
            data_format = self.data_format
            levels = self.levels
            ndims = self.ndims
        try:
            if data_format[1] == 'C':
                strides = [tf.cast(tf.math.floor(inp.shape[2+k]/levels[0][k]), dtype = tf.int32) for k in range(ndims)]
                windowshape = [strides[k] + (inp.shape[2+k] % levels[0][k]) for k in range(ndims)]
            else:
                strides = [tf.cast(tf.math.floor(inp.shape[2+k]/levels[0][k]), dtype = tf.int32) for k in range(ndims)]
                windowshape = [strides[k] + (inp.shape[1+k] % levels[0][k]) for k in range(ndims)]
            out = tf.keras.backend.reshape(tf.nn.pool(inp, window_shape=windowshape, strides=strides, pooling_type=self.pooling_type, padding = 'VALID', data_format=data_format), [inp.shape[0], -1])
            #out = tf.keras.backend.pool2d(inp, tuple(windowshape), padding = 'valid', data_format = 'channels_first', pool_mode = 'avg', strides = tuple(strides))

            for i in range(1,len(levels)):
                if data_format[1] == 'C':
                    strides = [tf.cast(tf.math.floor(inp.shape[2+k]/levels[i][k]), dtype = tf.int32) for k in range(ndims)]
                    windowshape = [strides[k] + (inp.shape[2+k] % levels[i][k]) for k in range(ndims)]
                else:
                    strides = [tf.cast(tf.math.floor(inp.shape[2+k]/levels[i][k]), dtype = tf.int32) for k in range(ndims)]
                    windowshape = [strides[k] + (inp.shape[1+k] % levels[i][k]) for k in range(ndims)]
                out = tf.concat([out, tf.reshape(tf.nn.pool(inp, window_shape=windowshape, strides=strides, pooling_type=self.pooling_type, padding = 'VALID', data_format=data_format), [inp.shape[0],-1])],1)
        except:
            print('spp init')
            if data_format[1] == 'C':
                out = tf.reduce_max(inp, axis = [k for k in range(2,2+ndims)])
            else:
                out = tf.reduce_max(inp, axis = [k for k in range(1,1+ndims)])
            levels = tf.constant(levels)
            out = tf.tile(out, [1,tf.reduce_sum(tf.reduce_prod(levels, axis = 1))])
        return out



import tensorflow as tf
import numpy as np
import opt_einsum as oe
import copy

from ..layers import SpatialPyramidPool
from .custom_blocks import ResnetBlock
from .Model_With_Integral_Loss import Model_With_Integral_Loss_ABC
from .Poisson_CNN import channels_first_flip_left_right

from ..dataset.generators.numerical import set_max_magnitude_in_batch as smmib
from ..misc import convolutional_poisson_loss

class Dirichlet_BC_NN_Series(Model_With_Integral_Loss_ABC):
    def __init__(self, data_format = 'channels_first', spp_levels = [[2],3,4,6,8,12,20,50], nmodes = 32, npolynomial = 5, conv_stages = 5, pooling_block_number = 5, final_channels = 10, initial_kernel_size = 19, final_kernel_size = 3, dense_layer_number = 5, initial_dense_layer_units = 250, kernel_regularizer = None, bias_regularizer = None, x_output_resolution = 256, convolutional_poisson_loss_stencil = 'five-point', **kwargs):
        super().__init__(**kwargs)
        self.data_format = data_format
        self.x_output_resolution = x_output_resolution
        self.nmodes = nmodes
        self.npolynomial = npolynomial
        self.modes_arange = tf.cast(tf.range(1,self.nmodes+1,delta=1), tf.keras.backend.floatx())
        
        if self.data_format == 'channels_first':
            self.space_dims_start = 2
        else:
            self.space_dims_start = 1

        
        spp_levels.append(None)
        self.conv0 = tf.keras.layers.Conv1D(filters = 2, kernel_size = initial_kernel_size, activation = tf.nn.leaky_relu, data_format = self.data_format, padding = 'same', kernel_regularizer = kernel_regularizer, bias_regularizer = bias_regularizer)
        self.pooling_blocks = []
        for j in range(pooling_block_number):
            conv_blocks = []
            for k in range(conv_stages):
                ksize = initial_kernel_size + (k*(final_kernel_size - initial_kernel_size)//(conv_stages-1))
                filters = final_channels-((k*final_channels)//(conv_stages))
                block = []
                if (k == 0) and (j != 0):
                    block.append(tf.keras.layers.AveragePooling1D(pool_size = 2**j, data_format = self.data_format))
                block.append(tf.keras.layers.Conv1D(filters = filters, kernel_size = ksize, activation=tf.nn.leaky_relu, data_format = self.data_format, padding = 'same', kernel_regularizer = kernel_regularizer, bias_regularizer = bias_regularizer))
                block.append(ResnetBlock(ndims = 1, data_format = self.data_format, filters = filters, kernel_size = ksize, activation=tf.nn.leaky_relu, kernel_regularizer = kernel_regularizer, bias_regularizer = bias_regularizer))
                if k == (conv_stages-1):
                    block.append(SpatialPyramidPool(spp_levels[:-(j+1)], data_format = self.data_format))
                conv_blocks.append(block)
            self.pooling_blocks.append(conv_blocks)

        self.dense_layers = []
        for k in range(dense_layer_number):
            units = initial_dense_layer_units + (k*(nmodes - initial_dense_layer_units)//(dense_layer_number-1))
            self.dense_layers.append(tf.keras.layers.Dense(units, activation = tf.nn.leaky_relu))
        #self.dense_layers.append(tf.keras.layers.Dense(nmodes+npolynomial*(npolynomial-1)/2, activation = tf.nn.leaky_relu))
        self.fourier_dense = tf.keras.layers.Dense(self.nmodes, activation = tf.nn.tanh) #26 sep  set activations to tanh
        if npolynomial != 0:
            self.poly_dense = tf.keras.layers.Dense(self.nmodes*self.npolynomial*(self.npolynomial+1)/2, activation = tf.nn.tanh)

    @staticmethod
    def build_polynomials(nx, order):
        order = order + 1
        result = []
        x = tf.cast(tf.keras.backend.arange(0,stop=nx)/(nx-1), tf.keras.backend.floatx())
        for k in range(order):
            if k == 0:
                result.append(tf.ones((nx), dtype = tf.keras.backend.floatx()))
            else:
                for j in range(k+1):
                    result.append((x**j)*((1-x)**(k-j)))
        return tf.stack(result)
            
            
    def call(self, inp):
        self.dx = inp[1]

        #set output x resolution
        try:
            if inp[2].shape[0] == 1: #set output shape
                self.x_output_resolution = int(inp[2])
            elif self.data_format == 'channels_first':
                self.x_output_resolution = int(inp[2][2])
            else:
                self.x_output_resolution = int(inp[2][1])
        except:
            pass

        #get domain info
        if self.data_format == 'channels_first': 
           domain_info = tf.einsum('ij,j->ij',tf.tile(inp[1], [1,3]), tf.stack([tf.constant(1.0, dtype = tf.keras.backend.floatx()), tf.cast(self.x_output_resolution, tf.keras.backend.floatx()), tf.cast(tf.constant(inp[0].shape[2]), tf.keras.backend.floatx())], axis = 0)) #calculate domain info
           input_length = inp[0].shape[2] #get mesh size in the direction of the provided BC
        else:
           domain_info = tf.einsum('ij,j->ij',tf.tile(inp[1], [1,3]), tf.stack([tf.constant(1.0, dtype = tf.keras.backend.floatx()), tf.cast(self.x_output_resolution, tf.keras.backend.floatx()), tf.cast(tf.constant(inp[0].shape[1]), tf.keras.backend.floatx())], axis = 0))
           input_length = inp[0].shape[1]

        #actual NN bit
        amplitudes = self.conv0(inp[0])
        pooling_block_results = []
        for pooling_block in self.pooling_blocks:
            pbr = amplitudes + 0.0
            for block in pooling_block:
                for layer in block:
                    pbr = layer(pbr)
            pooling_block_results.append(pbr)
        amplitudes = tf.keras.backend.concatenate(pooling_block_results + [domain_info], 1)

        #amplitudes = self.spp(amplitudes)
        #amplitudes = tf.keras.backend.concatenate([amplitudes,domain_info], 1)

        for layer in self.dense_layers:
            amplitudes = layer(amplitudes)

        if self.npolynomial != 0:
            poly_amplitudes = tf.keras.backend.reshape(self.poly_dense(amplitudes), (-1, self.nmodes, int(self.npolynomial*(self.npolynomial+1)/2)))
        amplitudes = self.fourier_dense(amplitudes)

        mode_coefficient_adjustment = tf.einsum('i,ij->ij',(2/domain_info[:,2])**2,1/tf.sinh(tf.einsum('i,j->ij',domain_info[:,1]/domain_info[:,2],-self.modes_arange*np.pi))) #26 sep - squared 2/domain_info[:,2]
        
        amplitudes = amplitudes * mode_coefficient_adjustment
        

        #build grids
        sinh = tf.sinh(tf.einsum('i,j,k->ijk', domain_info[:,1]/domain_info[:,2], self.modes_arange, -tf.keras.backend.arange(0,self.x_output_resolution, dtype = tf.keras.backend.floatx())/tf.cast(self.x_output_resolution-1, tf.keras.backend.floatx())*tf.constant(np.pi, dtype = tf.keras.backend.floatx())))
        sin = tf.sin(tf.einsum('j,l->jl',  self.modes_arange, tf.keras.backend.arange(0,input_length, dtype = tf.keras.backend.floatx())/tf.cast(input_length-1, tf.keras.backend.floatx()))*tf.constant(np.pi, dtype = tf.keras.backend.floatx()))

        if self.npolynomial != 0:
            poly_amplitudes = tf.einsum('ijm,ij->ijm', poly_amplitudes, amplitudes) #26 sep - changed 2nd einsum argument from mode coeff adjustment to amplitudes
            if self.data_format == 'channels_first':
                poly = tf.einsum('ijm,ml->ijl', poly_amplitudes, self.build_polynomials(tf.keras.backend.shape(inp[0])[-1], self.npolynomial-1))
            else:
                poly = tf.einsum('ijm,ml->ijl', poly_amplitudes, self.build_polynomials(tf.keras.backend.shape(inp[0])[1], self.npolynomial-1))
        else:
            poly = tf.keras.backend.zeros(list(tf.shape(amplitudes)) + [tf.shape(sin)[-1]], dtype = tf.keras.backend.floatx())
        
        #calculate result
        try:
            #out = oe.contract('ijl,ijk->ikl', poly, sinh, backend='tensorflow')
            out = oe.contract('ijl,ijk->ikl',oe.contract('ij,jl->ijl', amplitudes, sin, backend = 'tensorflow')+poly,sinh, backend='tensorflow')
            #out = oe.contract('ijk,jl,ij->ikl', sinh, sin, amplitudes, backend = 'tensorflow')
        except:
            #out = tf.einsum('ijl,ijk->ikl', poly, sinh)
            out = tf.einsum('ijl,ijk->ikl',tf.einsum('ij,jl->ijl', amplitudes, sin)+poly,sinh)
            #out = tf.einsum('ijk,jl,ij->ikl', sinh, sin, amplitudes)
            
        
        if self.data_format == 'channels_first':
            out = tf.expand_dims(out, axis = 1)
            return smmib(channels_first_flip_left_right(out),1.0)
        else:
            out = tf.expand_dims(out, axis = -1)

    
        

        
        
        

        



import numpy as np
import tensorflow as tf
import itertools, h5py
import opt_einsum as oe
from multiprocessing import Pool as ThreadPool

class integrator_nd:
    '''
    Class that integrates a given function f in an n-dimensional box.
    
    --Init arguments--
    -domain       : Defines the boundaries of the box. Format is [dim1 start, dim1 end, dim2 start, dim2 end, ..., dim_n start, dim_n end]
    -n_quadpts    : No of Gauss Legendre quadrature points across across each dimension
    
    --Call arguments--
    -f            : Callable taking n double-precision tensorflow tensors (shape of each: n_quadpts x n_quadpts x ... x n_quadpts, repeated n times) as arguments (n = dimensionality of the integration domain, i.e. len(domain)//2) and returning one double-precision tensorflow tensor of shape n_quadpts x n_quadpts x ... x n_quadpts, repeated n times. Represents the function to integrate. The inputs will be meshgrids containing the coordinates of the function to integrate, and output must be the values of the function.
    
    --Example--
    We want to integrate e^(-x^2-y^2-z^2) in the domain x = [1,2], y = [1,2], z = [1,2].
    First we need to define our function: g = lambda x,y,z: tf.exp(-x**2-y**2-z**2)
    Then we create the integreator: itg = integrator_nd(domain=[1.0,2.0,1.0,2.0,1.0,2.0])
    Finally, we evaluate the integral by calling the integrator: itg(g)
    '''
    
    def __init__(self, domain = [0,1,0,1], n_quadpts = 20):
        ndims = len(domain)//2
        quadrature_x, quadrature_w = tuple([np.polynomial.legendre.leggauss(n_quadpts)[i].astype(np.float64) for i in range(2)]) #quadrature weights and points
        c = tf.constant(np.array([np.array(0.5*(domain[n+1] - domain[n]),dtype=np.float64) for n in range(0,len(domain),2)]), dtype = tf.keras.backend.floatx()) #scaling coefficients - for handling domains other than [-1,1] x [-1,1]
        d = tf.constant(np.array([np.array(0.5*(domain[n+1] + domain[n]),dtype=np.float64) for n in range(0,len(domain),2)]), dtype = tf.keras.backend.floatx())
        self.quadpts = tf.constant(np.apply_along_axis(lambda x: x + d, 0, oe.contract('i...,i->i...',tf.stack(np.meshgrid(*list(itertools.repeat(quadrature_x,ndims)),indexing = 'xy'), axis=0),c, backend = 'tensorflow')).transpose(list(np.arange(1,ndims+1)) + [0]),dtype = tf.keras.backend.floatx())
    
        self.quadweights = tf.reduce_prod(c)*quadrature_w
    
        for i in range(1,ndims):
            self.quadweights = oe.contract('...,i->...i',self.quadweights,quadrature_w)
                                      
    def __call__(self, f):
        fi = f(*tf.unstack(self.quadpts, axis = -1))
        return tf.reduce_sum(tf.multiply(self.quadweights,fi))

def mode_coeff_calculation_multiprocessing_wrapper(args):
    #Integrates F(x1,x2,...,xn) * sin((m_1+1)*pi*x_1/L_1) * ... * sin((m_n+1)*pi*x_n/L_n)) dx_1 ... dx_n. Written as a separate function to permit multiprocessing pool map
    
    F = args[0]
    domain_volume = args[1]
    mplus1_pi_over_L = args[2]
    integrator = args[3]
    two_to_the_power_ndims = args[4]
    
    coefficients = []
    for i in range(int(mplus1_pi_over_L.shape[0])):
        integrand = lambda *vars: F(*vars) * tf.reduce_prod(tf.sin(oe.contract('i...,i->i...',tf.stack(vars),mplus1_pi_over_L[i],backend = 'tensorflow')),axis=0)
        coefficients.append(-integrator(integrand) * tf.cast(two_to_the_power_ndims,tf.keras.backend.floatx()) / (tf.cast(domain_volume,tf.keras.backend.floatx()) * tf.reduce_sum(tf.square(mplus1_pi_over_L[i]))))
    
    return tf.constant(np.array(coefficients), dtype = tf.keras.backend.floatx())

def homogeneous_analytical_dataset(rhs = 'random', output_shape = (64,64), nmodes = (16,16), domain = [1,1], n_threads = 1, return_rhs = True, max_magnitude = np.inf, batch_size = 1, expanded_dims = False, return_dx = False):
    '''
    Generates an analytical solution to the Poisson equation with homogeneous Dirichlet (i.e. 0) Boundary conditions in the box x_1 = [0, L_1], ..., x_n = [0, L_n]
    Solution strategy is Fourier series based as outlined in
    https://en.wikiversity.org/wiki/Partial_differential_equations/Poisson_Equation#Solution_to_Case_with_4_Homogeneous_Boundary_Conditions
    
    --Inputs--
    -rhs: Callable compatible with integrator_nd, or 'random'. If a callable is given, an analytical  solution for that function will be generated. If 'random', a random function containing random Fourier modes defined in nmodes will be returned.
    -output_shape: Shape of the output
    -nmodes: # of Fourier modes to include for each direction
    -domain: [L_1,L_2,...L_n]
    -n_threads: no of multiprocessing threads. useful only in non-random rhses
    -rhs_return: If set to true, it'll also return the RHS source term of the Poisson equation sampled at the same points as the solution
    -max_magnitude: If rhs is 'random', the solution will be scaled such that the maximum absolute value of the output will be this value.
    -expanded_dims: if set to True, the values returned will have an extra axis of length 1 as the axis #1 (not #0!)
    '''
    
    coords = [tf.linspace(tf.constant(0.0,dtype = tf.keras.backend.floatx()),domain[i], output_shape[i]) for i in range(len(output_shape))] #Evaluate coordinates along each axis
    coord_meshes = tf.stack(tf.meshgrid(*coords)) #Create meshgrid
    ndims = len(domain) #No of dims of function
    mode_permutations = tf.constant(np.array(list(itertools.product(*[np.arange(nmodes[i]) for i in range(len(nmodes))]))), dtype = tf.keras.backend.floatx()) #Every permutation of Fourier modes along different axes possible
    mplus1_pi_over_L = oe.contract('j,ij->ij',tf.constant(domain, dtype = tf.keras.backend.floatx())**(-1),(mode_permutations + 1)*np.pi, backend = 'tensorflow') #compute (m_i+1)*pi/L_i
    #print(mplus1_pi_over_L)
    #print(coord_meshes)
    sine_vals = tf.reduce_prod(tf.sin(oe.contract('ij,j...->ij...', mplus1_pi_over_L, coord_meshes, backend = 'tensorflow')),axis=1) #sin((m_1+1)*pi*x_1/L_1) * ... * sin((m_n+1)*pi*x_n/L_n))

    if rhs == 'random': #Generate random Fourier coefficients for a random RHS
        #For an n dimensional RHS function F(x_1,...x_n) = \sum_{m_1,...,m_n} (A_{m_1,...,m_n} * sin((m_1+1)*pi*x_1/L_1) * ... * sin((m_n+1)*pi*x_n/L_n))
        #\int_0^{L_n} ... \int_0^{L_1} F * sin((m_1+1)*pi*x_1/L_1) * ... * sin((m_n+1)*pi*x_n/L_n) dx_1 ... dx_n = A*L_1*...*L_n/2^n
        #Thus the solution to the Poisson equation \nabla^2 u = F will be u = a_{m_1,...,m_n} * sin((m_1+1)*pi*x_1/L_1) * ... * sin((m_n+1)*pi*x_n/L_n))
        #where a_{m_1,...,m_n} = -A_{m_1,...,m_n} * /(((m_1+1)*pi/L_1)^2+...+((m_n+1)*pi/L_n)^2)
        rhs_function_coeffs = oe.contract('ij,j->ij',2*tf.random.uniform(tf.stack([batch_size, mode_permutations.shape[0]]),dtype = tf.keras.backend.floatx())-1,tf.exp(-tf.reduce_sum(tf.cast(mode_permutations, tf.keras.backend.floatx()),axis=1)), backend = 'tensorflow') #Random RHS function Fourier coefficients
        soln_function_coeffs = -oe.contract('ij,j->ij',rhs_function_coeffs, tf.reduce_sum(tf.square(mplus1_pi_over_L),axis=1)**(-1), backend = 'tensorflow') #Random solution Fourier coefficients
        
        
        rhs = tf.Variable(oe.contract('ij,j...->i...', rhs_function_coeffs, sine_vals, backend = 'tensorflow'))
        soln = tf.Variable(oe.contract('ij,j...->i...', soln_function_coeffs, sine_vals, backend = 'tensorflow'))
        maxminratio = np.zeros((batch_size))
        
        if max_magnitude != np.inf:
            for i in range(int(rhs.shape[0])):
                scaling_factor = max_magnitude/tf.reduce_max(tf.abs(rhs[i,...]))
                maxminratio[i] = tf.abs(tf.reduce_max(rhs[i,...]) / tf.reduce_min(rhs[i,...]))
                rhs[i,...].assign(rhs[i,...] * scaling_factor)
                soln[i,...].assign(soln[i,...] * scaling_factor)

    elif callable(rhs):
        if np.prod(nmodes) % n_threads != 0:
            raise(ValueError('n_threads must divide reduce_prod(nmodes)'))
        
        pool = ThreadPool(n_threads) 
        full_domain = np.zeros((len(domain)*2)) #add lower bound 0s to domain argument for compatibility with integrate_nd
        full_domain[1::2] += np.array(domain)
        
        itg = integrator_nd(domain = full_domain)
        
        soln_function_coeffs = tf.concat(list(map(mode_coeff_calculation_multiprocessing_wrapper, zip(itertools.repeat(rhs,n_threads), itertools.repeat(tf.reduce_prod(domain),n_threads), tf.split(mplus1_pi_over_L, n_threads, axis=0), itertools.repeat(itg,n_threads), itertools.repeat(2**ndims,n_threads)))),axis=0)
        
        #soln_function_coeffs = tf.concat(pool.map(mode_coeff_calculation_multiprocessing_wrapper, zip(itertools.repeat(rhs,n_threads), itertools.repeat(tf.reduce_prod(domain),n_threads), tf.split(mplus1_pi_over_L, n_threads, axis=0), itertools.repeat(itg,n_threads), itertools.repeat(2**ndims,n_threads))),axis=0)

        soln = oe.contract('i,i...->...', soln_function_coeffs, sine_vals, backend = 'tensorflow')
        rhs = rhs(*coord_meshes)
    else:
        raise(TypeError('rhs must either be a callable function to be evaluated or random'))

    if expanded_dims:
        rhs = tf.expand_dims(rhs, axis = 1)
        soln = tf.expand_dims(soln, axis = 1)
    inp = []
    if return_rhs:
        inp.append(rhs)
    if return_dx:
        dx = [[domain[k]/(output_shape[k]-1) for k in range(len(domain))]]
        dx = tf.tile(tf.constant(dx, dtype = tf.keras.backend.floatx()), [batch_size, 1])
        inp.append(dx)
    return inp, soln
            
        
def random_calculation_multiprocessing_wrapper(args):
    

    
    batch_size = args[0]
    solution_generator_parameters = args[1]
    rhses = []
    solns = []
    for i in range(batch_size):
        rhs, soln = generate_analytical_solution_homogeneous_bc(**solution_generator_parameters)
        rhses.append(rhs)
        solns.append(solns)
    return np.concatenate(rhses), np.concatenate(solns)

class homogeneous_analytical_dataset_generator(tf.keras.utils.Sequence):
    def __init__(self, batch_size = 1, batches_per_epoch = 1, nmodes = 'random', domain = 'random', output_shape = 'random', randomize_rhs_max_magnitude = False, nmodes_random_range = [[8,24],[8,24]], random_output_shape_range = [[64,85],[64,85]], domain_random_range = [[0.5,1.5],[0.5,1.5]], rhs_random_max_magnitude = 1.0, keep_dx_constant = True, **analytical_dataset_arguments):
        '''
        Creates a generator that can be used to generate infinitely many sets of Poisson eq. RHSes-BCs-solutions
        
        batch_size: Int. Batch size of outputs.
        batches_per_epoch: Int. No of batches to create in each training epoch.
        randomize_rhs_smoothness: Boolean. Set to True if it's desired to have random RHS smoothnesses.
        rhs_random_smoothness_range: List of 2 integers. First integer is the lower bound and 2nd integer is the upper bound of the random RHS smoothnesses. Ignored if randomize_rhs_smoothness is False.
        randomize_boundary_smoothness: Boolean. Set to True if it's desired to have random BC smoothnesses.
        boundary_random_smoothness_range: Dict containing entries 'left', 'top', 'right', 'bottom'. Each entry must be the same format as rhs_random_smoothness_range
        randomize_rhs_max_magnitude: Boolean. Set to True if it's desired to have random RHS magnitudes.
        rhs_random_max_magnitude: Float. Max value of the random max magnitude.
        randomize_boundary_max_magnitude: Boolean. Set to True if it's desired to have random BC magnitudes.
        boundary_random_max_magnitude: Dict containing entries 'left', 'top', 'right', 'bottom'. Each entry must be the same format as rhs_random_max_magnitude.
        numerical_dataset_arguments: Arguments to pass onto the function numerical_dataset.
        return_keras_style: If set to True, the values from the boundaries dict will be unpacked (in the order left-top-right-bottom) into members of the output list.
        exclude_zero_boundaries: If set to True, nonzero boundaries are not returned if return_keras_style is True.
        '''

        self.nmodes = nmodes
        if self.nmodes == 'random':
            self.nmodes_random_range = nmodes_random_range
        self.domain = domain
        if self.domain == 'random':
            self.domain_random_range = domain_random_range
        self.output_shape = output_shape
        if self.output_shape == 'random':
            self.random_output_shape_range = random_output_shape_range
        self.randomize_rhs_max_magnitude = randomize_rhs_max_magnitude
        if self.randomize_rhs_max_magnitude:
            self.rhs_random_max_magnitude = rhs_random_max_magnitude

        self.batch_size = batch_size
        self.batches_per_epoch = batches_per_epoch
        self.ada = analytical_dataset_arguments
        self.keep_dx_constant = keep_dx_constant

    def __len__(self):
        return self.batches_per_epoch

    def __getitem__(self, idx): #Generates a batch. Input idx is ignored but is necessary per Keras API.

        if self.randomize_rhs_max_magnitude:
            self.ada['rhs_max_magnitude'] = np.random.rand() * self.rhs_random_max_magnitude

        if self.nmodes == 'random':
            self.ada['nmodes'] = tuple([int(np.random.randint(self.nmodes_random_range[k][0], high = self.nmodes_random_range[k][1])) for k in range(len(self.nmodes_random_range))])
        else:
            self.ada['nmodes'] = self.nmodes

        if self.output_shape == 'random':
            self.ada['output_shape'] = tuple([int(np.random.randint(self.random_output_shape_range[k][0], high = self.random_output_shape_range[k][1])) for k in range(len(self.random_output_shape_range))])
        else:
            self.ada['output_shape'] = self.output_shape

        if self.domain == 'random' and self.keep_dx_constant:
            self.ada['domain'] = [np.random.rand()*(self.domain_random_range[0][1] - self.domain_random_range[0][0]) + self.domain_random_range[0][0]]
            dx = self.ada['domain'][0]/(self.ada['output_shape'][0] - 1)
            for k in range(1, len(self.domain_random_range)):
                self.ada['domain'].append(dx * (self.ada['output_shape'][k] - 1))
        elif self.domain == 'random':
            self.ada['domain'] = [np.random.rand()*(self.domain_random_range[k][1] - self.domain_random_range[k][0]) + self.domain_random_range[k][0] for k in range(len(self.domain_random_range))]
        
            
        inp, out = homogeneous_analytical_dataset(**self.ada, batch_size = self.batch_size)

        if self.keep_dx_constant:
            inp[1] = tf.expand_dims(inp[1][:,0], axis = 1)
        return inp, out
        




import tensorflow as tf
from tensorflow.keras.layers import Input, Conv2D
from tensorflow.keras.models import Model
import opt_einsum as oe
import numpy as np

class convolutional_poisson_loss():
    def __init__(self, h = 0.05, compressible = False, stencil = 'five-point', return_pointwise = False, data_format = 'channels_first'):
        self.return_pointwise = return_pointwise
        self.compressible = compressible
        self.data_format = data_format
        inp_soln = Input((1, None, None))
        inp_dx = Input((1,))
        if compressible == False:
            if stencil == 'nine-point':
                w = np.zeros((5,5,1,1), dtype = tf.keras.backend.floatx())
                w[2,0,0,0] = -1
                w[0,2,0,0] = -1
                w[1,2,0,0] = 16.0
                w[2,1,0,0] = 16.0
                w[2,2,0,0] = -60
                w[3,2,0,0] = 16.0
                w[2,3,0,0] = 16.0
                w[2,4,0,0] = -1
                w[4,2,0,0] = -1
                w = w/12.0
                kernel_size = 5
            elif stencil == 'five-point':
                w = np.zeros((3,3,1,1), dtype = tf.keras.backend.floatx())
                w[1,1,0,0] = -4.0
                w[0,1,0,0] = 1.0
                w[1,0,0,0] = 1.0
                w[2,1,0,0] = 1.0
                w[1,2,0,0] = 1.0
                kernel_size = 3
            laplacian_kernel = Conv2D(filters=1, kernel_size=kernel_size, activation='linear', data_format = data_format, padding = 'valid', use_bias = False)(inp_soln)
            def f(x):
                try:
                    return oe.contract('i...,i->i...', x[0], tf.squeeze(1/(x[1]**2), axis = 1))
                except:
                    return x[0]
            dx_adjustment = tf.keras.layers.Lambda(f)([laplacian_kernel, inp_dx])
            mod = Model([inp_soln, inp_dx],laplacian_kernel)
            mod.set_weights([tf.constant(w, dtype=tf.keras.backend.floatx())])
        else:
            if stencil == 'nine-point':
                w = np.zeros((5,5,1,2), dtype = tf.keras.backend.floatx())
                w[0,2,0,0] = 1.0
                w[1,2,0,0] = -8.0
                w[2,2,0,0] = 0.0
                w[3,2,0,0] = 8.0
                w[4,2,0,0] = -1.0
                w = w/12.0
                kernel_size = 5
            if stencil == 'five-point':
                w = np.zeros((3,3,1,2), dtype = tf.keras.backend.floatx())
                w[0,1,0,0] = -1.0
                w[1,1,0,0] = 0.0
                w[2,1,0,0] = 1.0
                w = w/2.0
                kernel_size = 3
            w[...,0,1] = w[...,0,0].transpose()
            inp_rho = Input((1, None, None))
            grad_u = Conv2D(filters = 2, kernel_size = kernel_size, activation = 'linear', data_format = data_format, padding = 'valid', use_bias = False)(inp_soln)
            def f(x):
                try:
                    return oe.contract('i...,i->i...', x[0], tf.squeeze(1/(2*x[1]), axis = 1), backend = 'tensorflow')
                except:
                    return x[0]
            dx_adjustment_1 = tf.keras.layers.Lambda(f)([grad_u, inp_dx])
            if data_format == 'channels_first':
                def g(x):
                    try:
                        return oe.contract('...,...->...', x[0], x[1][...,kernel_size//2:-kernel_size//2,kernel_size//2:-kernel_size//2], backend = 'tensorflow')
                    except:
                        return x[0]
                rho_adjustment = tf.keras.layers.Lambda(g)([dx_adjustment_1, inp_rho])
            else:
                def g(x):
                    try:
                        return oe.contract('...,...->...', x[0], x[1][:,kernel_size//2:-kernel_size//2,kernel_size//2:-kernel_size//2,:], backend = 'tensorflow')
                    except:
                        return x[0]
                rho_adjustment = tf.keras.layers.Lambda(g)([dx_adjustment_1, inp_rho])                                 
            div = Conv2D(filters = 1, kernel_size = kernel_size, activation = 'linear', data_format = data_format, padding = 'valid', use_bias = False)(rho_adjustment)
            mod = Model([inp_soln, inp_rho, inp_dx], div)
            mod.set_weights([w, w.transpose((0,1,3,2))])
        self.mod = mod
    def evaluate_loss(self, soln, rhs, dx, rho = None):
        if self.compressible and (rho is None):
            raise(ValueError('rho must be provided for the compressible case'))
        if (not self.compressible) and (rho is not None):
            print('Warning: rho has been provided for an incompressible instance, and will be disregarded.')

        if isinstance(dx, float):
            dx = dx * tf.ones((soln.shape[0],1), dtype = tf.keras.backend.floatx())

        if self.compressible:
            out = self.mod([soln, rho, dx])
        else:
            out = self.mod([soln, dx])

        if self.data_format == 'channels_first':
            out = (out - rhs[...,(rhs.shape[-2] - out.shape[-2])//2:-(rhs.shape[-2] - out.shape[-2])//2,(rhs.shape[-1] - out.shape[-1])//2:-(rhs.shape[-1] - out.shape[-1])//2])**2
        else:
            out = (out - rhs[:,(rhs.shape[-3] - out.shape[-3])//2:-(rhs.shape[-3] - out.shape[-3])//2,(rhs.shape[-2] - out.shape[-2])//2:-(rhs.shape[-2] - out.shape[-2])//2,:])**2
        if self.return_pointwise:
            return out
        else:
            return tf.reduce_mean(out)

# def conv_laplacian_loss(h, return_pointwise = False):
#     inp = Input((1,None, None))
#     laplacian_kernel = Conv2D(filters=1, kernel_size=5, activation='linear', data_format='channels_first', padding = 'same')(inp)
#     mod = Model(inp,laplacian_kernel)
#     w = np.zeros((5,5,1,1), dtype = np.float64)
#     w[2,0,0,0] = -1
#     w[0,2,0,0] = -1
#     w[1,2,0,0] = 16.0
#     w[2,1,0,0] = 16.0
#     w[2,2,0,0] = -60
#     w[3,2,0,0] = 16.0
#     w[2,3,0,0] = 16.0
#     w[2,4,0,0] = -1
#     w[4,2,0,0] = -1
#     mod.set_weights([(1/(12*h**2))*tf.constant(w, dtype=tf.float64),tf.constant([0.0], dtype=tf.float64)])
#     #pdb.set_trace()
#     if return_pointwise:
#         if int(tf.__version__[0]) < 2:
#             import tensorflow.contrib.eager as tfe
#             @tf.contrib.eager.defun
#             def laplacian_loss(rhs, solution):
#                 return (mod(solution)[:,:,2:-2,2:-2]-rhs[:,:,2:-2,2:-2])**2
#         else: 
#             @tf.function
#             def laplacian_loss(rhs, solution):
#                 return (mod(solution)[:,:,2:-2,2:-2]-rhs[:,:,2:-2,2:-2])**2
#     else:
#         if int(tf.__version__[0]) < 2:
#             @tf.contrib.eager.defun
#             def laplacian_loss(rhs, solution):
#                 return tf.reduce_sum((mod(solution)[:,:,2:-2,2:-2]-rhs[:,:,2:-2,2:-2])**2)/tf.cast(tf.reduce_prod(rhs[:,:,2:-2,2:-2].shape), rhs.dtype)
#         else:
#             @tf.function
#             def laplacian_loss(rhs, solution):
#                 return tf.reduce_sum((mod(solution)[:,:,2:-2,2:-2]-rhs[:,:,2:-2,2:-2])**2)/tf.cast(tf.reduce_prod(rhs[:,:,2:-2,2:-2].shape), rhs.dtype)
#     return laplacian_loss

# def compressible_poisson_loss(dx, dy = None, return_pointwise = False, data_format = data_format):
#     inp = Input((1,-1,-1))



import tensorflow as tf
import itertools
import opt_einsum as oe
import numpy as np
import copy

from ..layers import Upsample, DeconvUpscale2D, SpatialPyramidPool

class SepConvBlock(tf.keras.models.Model):
    def __init__(self, data_format = 'channels_first', separable_kernel_size  = (5,256), nonsep_kernel_size = 5, separable_activation = tf.nn.leaky_relu, nonsep_activation = tf.nn.leaky_relu, separable_filters = 8, nonsep_filters = 4):
        super().__init__()
        self.separableconv2d = tf.keras.layers.SeparableConv2D(separable_filters, kernel_size = separable_kernel_size, padding = 'same', activation = separable_activation, data_format = data_format)
        self.conv2d = tf.keras.layers.Conv2D(filters = nonsep_filters, kernel_size = nonsep_kernel_size, activation = nonsep_activation, padding = 'same', data_format = data_format)
        
    def call(self, inp):
        return self.conv2d(self.separableconv2d(inp))
    
class ResnetBlock(tf.keras.models.Model):
    def __init__(self, ndims = 2, data_format = 'channels_first', **conv_layer_args):
        super().__init__()
        if int(ndims) == 1:
            self.conv0 = tf.keras.layers.Conv1D(padding = 'same', data_format = data_format, **conv_layer_args)
            self.conv1 = tf.keras.layers.Conv1D(padding = 'same', data_format = data_format, **conv_layer_args)
        elif int(ndims) == 2:
            self.conv0 = tf.keras.layers.Conv2D(padding = 'same', data_format = data_format, **conv_layer_args)
            self.conv1 = tf.keras.layers.Conv2D(padding = 'same', data_format = data_format, **conv_layer_args)
        elif int(ndims) == 3:
            self.conv0 = tf.keras.layers.Conv3D(padding = 'same', data_format = data_format, **conv_layer_args)
            self.conv1 = tf.keras.layers.Conv3D(padding = 'same', data_format = data_format, **conv_layer_args)
        else:
            raise ValueError('ndims must be 1,2 or 3')

    def call(self, inp):
        return inp + self.conv0(self.conv1(inp))
    
class ResampledConvolutionBlock(tf.keras.models.Model):
    def __init__(self, resample_arg, data_format = 'channels_first', resize_method = tf.image.ResizeMethod.BICUBIC, use_resnetblocks = True, pre_upsample_conv_blocks = 2, post_upsample_conv_blocks = 2, **convblockargs):
        super().__init__()
        self.data_format = data_format
        self.resample = Upsample(resample_arg, data_format = data_format, resize_method = resize_method)
        self.resize_to_original_shape = Upsample([-1, -1], data_format = data_format)
        if use_resnetblocks:
            self.pre_upsample_conv_blocks = [ResnetBlock(data_format = data_format, **convblockargs) for i in range(pre_upsample_conv_blocks)]
            self.post_upsample_conv_blocks = [ResnetBlock(data_format = data_format, **convblockargs) for i in range(post_upsample_conv_blocks)]
        else:
            self.pre_upsample_conv_blocks = [tf.keras.layers.Conv2D(data_format = data_format, padding = 'same', **convblockargs) for i in range(pre_upsample_conv_blocks)]
            self.post_upsample_conv_blocks = [tf.keras.layers.Conv2D(data_format = data_format, padding = 'same', **convblockargs) for i in range(post_upsample_conv_blocks)]
        
    def call(self, inp):
        if self.data_format == 'channels_first':
            original_shape = [inp.shape[2], inp.shape[3]]
        else:
            original_shape = [inp.shape[1], inp.shape[2]]
        out = self.resample(inp)
        for convblock in self.pre_upsample_conv_blocks:
            out = convblock(out)
        out = self.resize_to_original_shape([out, original_shape])
        for convblock in self.post_upsample_conv_blocks:
            out = convblock(out)
        return out

    
    
class AveragePoolingBlock(tf.keras.models.Model):
    def __init__(self, pool_size = 2, data_format = 'channels_first', resize_method = tf.image.ResizeMethod.BICUBIC, use_resnetblocks = False, use_deconv_upsample = False, **convblockargs):
        super().__init__()
        self.data_format = data_format
        self.pool = tf.keras.layers.AveragePooling2D(data_format = data_format, pool_size = pool_size, padding = 'same')
        self.use_deconv_upsample = use_deconv_upsample
        
        if not use_deconv_upsample:
            self.upsample = Upsample([-1,-1],resize_method=resize_method, data_format = data_format)
        else:
            deconvblockargs = copy.deepcopy(convblockargs)
            #_ = deconvblockargs.pop('kernel_regularizer', None) #backprop has difficulties with regularizing deconv weights so remove these.
            #_ = deconvblockargs.pop('bias_regularizer', None)
            self.upsample = DeconvUpscale2D(pool_size, data_format = data_format, **deconvblockargs)

        if not use_resnetblocks:
            self.pooledconv = tf.keras.layers.Conv2D(data_format = data_format, padding='same', **convblockargs)
            self.upsampledconv = tf.keras.layers.Conv2D(data_format = data_format, padding='same', **convblockargs)
        else:
            self.pooledconv = ResnetBlock(data_format = data_format, **convblockargs)
            self.upsampledconv = ResnetBlock(data_format = data_format, **convblockargs)

    def call(self, inp):
        if self.data_format == 'channels_first':
            input_shape = [inp.shape[-2], inp.shape[-1]]
        else:
            input_shape = [inp.shape[-3],inp.shape[-2]]
        return self.upsampledconv(self.upsample([self.pooledconv(self.pool(inp)), input_shape]))
        

class Scaling(tf.keras.models.Model):
    def __init__(self, stages = 2, downsampling_ratio_per_stage = 2, data_format = 'channels_first', padding = 'same', spp_levels = [[2,2], 3, 5], **convargs):
        super().__init__()
        self.spp = SpatialPyramidPool(spp_levels, data_format = data_format, pooling_type = 'MAX')
        self.stages = []
        self.data_format = data_format
        for k in range(stages):
            self.stages.append(tf.keras.layers.Conv2D(data_format = data_format, padding = padding, **convargs))
            self.stages.append(tf.keras.layers.AveragePooling2D(pool_size = downsampling_ratio_per_stage, padding = padding, data_format = data_format))

        self.dense_0 = tf.keras.layers.Dense(100, activation = tf.nn.leaky_relu)
        self.dense_1 = tf.keras.layers.Dense(25, activation = tf.nn.leaky_relu)
        self.dense_2 = tf.keras.layers.Dense(1, activation = 'linear')

    def call(self, inp):
        #input format: tf.Tensor of shape [batch_size, 2, nx, ny] where [:,0,...] is the input to conditionally scale, or a list of 2 tf.Tensors such that [input_to_scale, 2nd input]
        if self.data_format == 'channels_first' and len(inp) == 3:
            conv_data = tf.concat(inp[:2], axis = 1)
            inp_to_scale = tf.expand_dims(conv_data[:,0,...], axis = 1)
        elif self.data_format == 'channels_first':
            conv_data = inp[0]
            inp_to_scale = tf.expand_dims(inp[0][:,0,...], axis = 1)
        elif len(inp) == 3:
            conv_data = tf.concat(inp[:2], axis = -1)
            inp_to_scale = tf.expand_dims(conv_data[:,...,0], axis = -1)
        else:
            conv_data = inp[0]
            inp_to_scale = tf.expand_dims(inp[:,...,0], axis = -1)
        out = self.stages[0](conv_data)
        for k in range(1, len(self.stages)):
            out = self.stages[k](out)
        out = self.spp(out)
        #out = tf.concat([out, inp[1]], axis = 1)
        out = self.dense_2(self.dense_1(self.dense_0(out)))
        return tf.einsum('i...,i...->i...', 1.0+out, inp_to_scale)
        
        
    



import tensorflow as tf
import copy

from .Poisson_CNN import Poisson_CNN
from .custom_blocks import ResnetBlock

from ..layers import SpatialPyramidPool
from ..dataset.generators.numerical import numerical_dataset_generator as ndg

class Poisson_Discriminator(tf.keras.models.Model):
    def __init__(self, data_format = 'channels_first', n_convstages = 3, conv_params = {'filters': 6, 'kernel_size': 13, 'activation': tf.nn.leaky_relu, 'padding': 'same', 'strides': 1}, resnetblock_params = {'kernel_size' : 13, 'activation' : tf.nn.leaky_relu}, spp_levels = [[3,3],4,6,8,12], dropout_fraction = 0.2, **kwargs):
        super().__init__(**kwargs)
        conv_params = copy.deepcopy(conv_params)
        self.data_format = data_format
        self.n_convstages = n_convstages
        if self.data_format == 'channels_first':
            batchnorm_axis = 1
        else:
            batchnorm_axis = -1
            
        stridedconv_filters = conv_params.pop('filters')
        stridedconv_filters = [stridedconv_filters*(2**(k)) for k in range(n_convstages)]
        self.reshaping_convs = [tf.keras.layers.Conv2D(filters = stridedconv_filters[k], data_format = self.data_format, **conv_params) for k in range(n_convstages)]
        self.resnet_blocks = [ResnetBlock(data_format = self.data_format, filters = stridedconv_filters[k], **resnetblock_params) for k in range(n_convstages)]
        self.batchnorms = [tf.keras.layers.BatchNormalization(axis = batchnorm_axis) for k in range(2*n_convstages)]
        self.rhs_spp = SpatialPyramidPool(spp_levels, data_format = self.data_format)
        
        self.dense_0 = tf.keras.layers.Dense(256, activation = tf.nn.leaky_relu)
        self.dense_1 = tf.keras.layers.Dense(100, activation = tf.nn.leaky_relu)
        self.dense_2 = tf.keras.layers.Dense(1, activation = tf.nn.sigmoid)
    def call(self, inp):
        out = self.batchnorms[1](self.resnet_blocks[0](self.batchnorms[0](self.reshaping_convs[0](inp))))
        for k in range(1,self.n_convstages):
            out = self.batchnorms[2*k+1](self.resnet_blocks[k](self.batchnorms[2*k](self.reshaping_convs[k](out))))
        out = self.rhs_spp(out)
        return self.dense_2(self.dense_1(self.dense_0(out)))

class combined_model(tf.keras.models.Model):
    def __init__(self, discriminator, generator):
        super().__init__()
        self.discriminator = discriminator
        self.discriminator.trainable = False
        self.generator = generator
    def call(self, inp):
        out = self.generator(inp)
        out = self.discriminator(out)
        return out#self.discriminator(self.generator(inp))
    
class Poisson_GAN():
    def __init__(self, Poisson_CNN_args = {}, data_format = 'channels_first', discriminator_weights = None, generator_weights = None, learning_rate = 1e-5):
        self.data_format = data_format
        self.optimizer =  tf.keras.optimizers.Adam(learning_rate = learning_rate)
        
        self.generator = Poisson_CNN(**Poisson_CNN_args)
        self.generator([tf.random.uniform((10,1,64,64), dtype = tf.keras.backend.floatx()), tf.random.uniform((10,1,64), dtype = tf.keras.backend.floatx()), tf.random.uniform((10,1,64), dtype = tf.keras.backend.floatx()), tf.random.uniform((10,1,64), dtype = tf.keras.backend.floatx()), tf.random.uniform((10,1,64), dtype = tf.keras.backend.floatx()), tf.random.uniform((10,1), dtype = tf.keras.backend.floatx())])
        if generator_weights is not None:
            self.generator.load_weights(generator_weights)
        self.generator.compile(loss = self.generator.integral_loss, optimizer = self.optimizer)
        self.generator.run_eagerly = True
        
        self.discriminator = Poisson_Discriminator()
        self.discriminator(tf.random.uniform((10,1,100,100), dtype = tf.keras.backend.floatx()))
        if discriminator_weights is not None:
            self.discriminator.load_weights(discriminator_weights)
        self.discriminator.compile(loss = 'binary_crossentropy', optimizer = self.optimizer)
        self.discriminator.run_eagerly = True
        
        # self.combined_model = tf.keras.models.Model(self.generator, self.discriminator)
        # self.combined_model = tf.keras.Sequential()
        # self.combined_model.add(self.generator)
        # self.combined_model.add(self.discriminator)
        self.combined_model = combined_model(self.discriminator, self.generator)
        self.combined_model.compile(loss = 'binary_crossentropy', optimizer = self.optimizer)
        self.combined_model.run_eagerly = True

    def train(self, epochs, dataset_generator = ndg(batch_size = 20, batches_per_epoch = 50, randomize_rhs_smoothness = True, randomize_boundary_smoothness = True, randomize_boundary_max_magnitudes = True, return_boundaries = True, return_dx = True, random_output_shape_range = [[256,384],[256,384]], random_dx_range = [0.005,0.02])):
        ###Implement weight saving!!!
        for epoch in range(epochs):
            '''
            Train discriminator
            '''
            inp, soln = dataset_generator.__getitem__(0)
            self.discriminator.trainable = True
            discriminator_loss_real = self.discriminator.train_on_batch(soln, tf.ones((soln.shape[0],1), dtype = tf.keras.backend.floatx()))
            generator_soln = self.generator.predict(inp)
            discriminator_loss_fake = self.discriminator.train_on_batch(generator_soln, tf.zeros((soln.shape[0],1), dtype = tf.keras.backend.floatx()))

            '''
            Train generator
            '''
            inp, soln = dataset_generator.__getitem__(0)
            self.combined_model.discriminator.trainable = False
            generator_loss = self.combined_model.train_on_batch(inp, tf.ones((soln.shape[0],1), dtype = tf.keras.backend.floatx()))

            '''
            Evaluate generator MSE
            '''
            generator_mse = tf.keras.losses.MSE(soln, self.combined_model.generator(inp))

            if epoch % 5 == 0:
                print('----Epoch ' + str(epoch) + '---- ')
                print('Discriminator loss on real inputs: ' + str(discriminator_loss_real))
                print('Discriminator loss on fake inputs: ' + str(discriminator_loss_fake))
                print('Combined model loss: ' + str(generator_loss))
                print('Generator MSE: ' + str(tf.reduce_mean(generator_mse).numpy()))
        



import tensorflow as tf
import copy

class DeconvUpscale2D(tf.keras.layers.Layer):
    def __init__(self, upsample_ratio, kernel_size, activation = tf.nn.leaky_relu, use_bias = True, kernel_initializer = 'glorot_uniform', bias_initializer = 'zeros', kernel_regularizer = None, bias_regularizer = None, data_format = 'channels_first', kernel_constraint = None, bias_constraint = None, input_channels = None, filters = None, **kwargs):

        self.data_format = data_format
        # if data_format == 'channels_first':
        #     self.data_format = 'NCHW'
        # else:
        #     self.data_format = 'NHWC'

        self.kernel_initializer = kernel_initializer
        self.bias_initializer = bias_initializer

        self.kernel_regularizer = kernel_regularizer
        self.bias_regularizer = bias_regularizer

        self.kernel_constraint = kernel_constraint
        self.bias_constraint = bias_constraint
        
        self.use_bias = use_bias
        self.activation = activation

        self.kernel_size = kernel_size
        if isinstance(self.kernel_size, int):
            self.kernel_size = [self.kernel_size for k in range(2)]
        
        self.upsample_ratio = upsample_ratio
        
        self.built = False
        
        super().__init__(**kwargs)

    def build(self, input_shape):
        if self.data_format == 'channels_first':#if self.data_format[1] == 'C':
            input_channels = input_shape[1]
        else:
            input_channels = input_shape[-1]
        
        self.kernel = self.add_weight(name = 'kernel', shape = tuple(list(self.kernel_size) + [input_channels, input_channels]), trainable = True, dtype = tf.keras.backend.floatx(), initializer = self.kernel_initializer, regularizer = self.kernel_regularizer, constraint = self.kernel_constraint)
        self.bias = self.add_weight(name = 'bias', shape = (input_channels,), trainable = True, dtype = tf.keras.backend.floatx(), initializer = self.bias_initializer, regularizer = self.bias_regularizer, constraint = self.bias_constraint)

        super().build(input_shape)

    def call(self, inp):
        output_shape = list(inp[1])
        if len(output_shape) != len(inp[0].shape):
            if self.data_format == 'channels_first':#if self.data_format[1] == 'C':
                output_shape = list(inp[0].shape)[:2] + output_shape
            else:
                input_shape = list(inp[0].shape)
                output_shape = [input_shape[0]] + output_shape + [input_shape[-1]]
        input_tensor = inp[0]
        #out = tf.nn.conv2d_transpose(inp[0], strides = self.upsample_ratio, filters = self.kernel, data_format = self.data_format, output_shape = output_shape, padding = 'SAME')
        #out = tf.nn.bias_add(out, self.bias, data_format = self.data_format)
        out = tf.keras.backend.conv2d_transpose(inp[0], self.kernel, strides = (self.upsample_ratio, self.upsample_ratio), data_format = self.data_format, output_shape = output_shape, padding = 'same')
        out = tf.keras.backend.bias_add(out, self.bias, data_format = self.data_format)
        return self.activation(out)
    
    def __call__(self, inp):
        if self.built == False:
            self.built = True
            self.build(inp[0].shape)
        return super().__call__(inp)
